import copy
import logging

from ... import emulators, exceptions, hinting, state
from .. import analysis

logger = logging.getLogger(__name__)


class CodeReachable(analysis.Analysis):
    """A simple analysis that logs what code is reachable by symbolic execution."""

    def __init__(self, *args, max_steps=500, **kwargs):
        self.steps_left = max_steps
        super().__init__(**kwargs)

    name = "code-reachable"
    description = ""
    version = "0.0.1"

    def run(self, state: state.Machine) -> None:
        machine = copy.deepcopy(state)
        cpu = machine.get_cpu()
        emulator = emulators.AngrEmulator(cpu.platform)
        machine.apply(emulator)

        try:
            while self.steps_left is None or self.steps_left > 0:
                emulator.step()
                if emulator.mgr:
                    for s in emulator.mgr.active:
                        pc = s._ip.concrete_value
                        hint = hinting.ReachableCodeHint(
                            message=f"Address {hex(pc)} is reachable via symbolic execution",
                            address=pc,
                        )
                        self.hinter.send(hint)
                if self.steps_left is not None:
                    self.steps_left -= 1
        except exceptions.EmulationStop:
            return
        except emulators.angr.PathTerminationSignal:
            return



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import AArch64SysVModel


class AArch64SysVMemcpy(Memcpy, AArch64SysVModel):
    pass


class AArch64SysVMemmove(Memmove, AArch64SysVModel):
    pass


class AArch64SysVStrcpy(Strcpy, AArch64SysVModel):
    pass


class AArch64SysVStrncpy(Strncpy, AArch64SysVModel):
    pass


class AArch64SysVStrcat(Strcat, AArch64SysVModel):
    pass


class AArch64SysVStrncat(Strncat, AArch64SysVModel):
    pass


class AArch64SysVMemcmp(Memcmp, AArch64SysVModel):
    pass


class AArch64SysVStrncmp(Strncmp, AArch64SysVModel):
    pass


class AArch64SysVStrcmp(Strcmp, AArch64SysVModel):
    pass


class AArch64SysVStrcoll(Strcoll, AArch64SysVModel):
    pass


class AArch64SysVStrxfrm(Strxfrm, AArch64SysVModel):
    pass


class AArch64SysVMemchr(Memchr, AArch64SysVModel):
    pass


class AArch64SysVStrchr(Strchr, AArch64SysVModel):
    pass


class AArch64SysVStrcspn(Strcspn, AArch64SysVModel):
    pass


class AArch64SysVStrpbrk(Strpbrk, AArch64SysVModel):
    pass


class AArch64SysVStrrchr(Strrchr, AArch64SysVModel):
    pass


class AArch64SysVStrspn(Strspn, AArch64SysVModel):
    pass


class AArch64SysVStrstr(Strstr, AArch64SysVModel):
    pass


class AArch64SysVStrtok(Strtok, AArch64SysVModel):
    pass


class AArch64SysVMemset(Memset, AArch64SysVModel):
    pass


class AArch64SysVStrerror(Strerror, AArch64SysVModel):
    pass


class AArch64SysVStrlen(Strlen, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVMemcpy",
    "AArch64SysVMemmove",
    "AArch64SysVStrcpy",
    "AArch64SysVStrncpy",
    "AArch64SysVStrcat",
    "AArch64SysVStrncat",
    "AArch64SysVMemcmp",
    "AArch64SysVStrncmp",
    "AArch64SysVStrcmp",
    "AArch64SysVStrcoll",
    "AArch64SysVStrxfrm",
    "AArch64SysVMemchr",
    "AArch64SysVStrchr",
    "AArch64SysVStrcspn",
    "AArch64SysVStrpbrk",
    "AArch64SysVStrrchr",
    "AArch64SysVStrspn",
    "AArch64SysVStrstr",
    "AArch64SysVStrtok",
    "AArch64SysVMemset",
    "AArch64SysVStrerror",
    "AArch64SysVStrlen",
]



from importlib import metadata as __metadata

metadata = __metadata.metadata("smallworld-re")

__title__ = metadata["name"]
__description__ = metadata["Summary"]
__author__ = metadata["Author"]
__version__ = metadata["version"]


from . import (
    analyses,
    emulators,
    exceptions,
    extern,
    hinting,
    instructions,
    logging,
    platforms,
    state,
)
from .helpers import *  # noqa: F401, F403
from .helpers import __all__ as __helpers__

__all__ = __helpers__ + [
    "analyses",
    "emulators",
    "exceptions",
    "extern",
    "hinting",
    "instructions",
    "logging",
    "platforms",
    "state",
]



import archinfo

from ....platforms import Architecture, Byteorder
from .machdef import AngrMachineDef


class ARMMachineDef(AngrMachineDef):
    def __init__(self):
        self._registers = {
            # *** General-purpose registers ***
            "r0": "r0",
            "r1": "r1",
            "r2": "r2",
            "r3": "r3",
            "r4": "r4",
            "r5": "r5",
            "r6": "r6",
            "r7": "r7",
            "r8": "r8",
            # r9 doubles as the Static base pointer
            "r9": "r9",
            "sb": "sb",
            # r10 doubles as the Stack Limit pointer
            "r10": "r10",
            "sl": "sl",
            # r11 doubles as the Frame Pointer, if desired.
            "r11": "r11",
            "fp": "fp",
            # r12 doubles as the Intra-call scratch register
            "r12": "r12",
            "ip": "ip",
            "sp": "sp",
            "lr": "lr",
            "pc": "pc",
        }


class ARMMachineMixinM:
    """Mixin class for M-series CPUs.

    The main difference between M and R/A
    is the available system status registers.
    """

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Special Registers ***
                # Program Status Register
                # NOTE: PSR can be accessed through several masked aliases.
                # These are read-only, so I'm not including them.
                # - apsr: Just the condition flags
                # - ipsr: Just exception information
                # - epsr: Just execution state info
                # - iapsr: apsr | ipsr
                # - eapsr: apsr | epsr
                # - iepsr: ipsr | epsr
                # - xpsr: apsr | ipsr | epsr
                "psr": "cpsr",
                # Exception Mask Register
                "primask": "primask",
                # Base Priority Mask Register
                "basepri": "basepri",
                # Fault Mask Register
                "faultmask": "faultmask",
                # Control register; includes a lot of flags.
                "control": "control",
                # *** Stack Pointer Bank ***
                # sp is actually an alias to one of these two.
                # Exactly which one depends on a bit in control.
                # Emulators that care should be careful when loading state.
                # Main Stack Pointer
                "msp": "msp",
                # Process Stack Pointer
                "psp": "psp",
            }
        )


class ARMMachineMixinRA:
    """Mixin for ARM series A and R CPUs"""

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Special Registers ***
                # Current Program Status Register
                # NOTE: CPSR can be accessed through several masked aliases.
                # These are read-only, so I'm not including them.
                # - isetstate: Just includes instruction set control bits
                # - itstate: Just includes state bits for Thumb IT instruction
                "cpsr": "cpsr",
                # Saved Program Status Register
                "spsr": "spsr",
                # *** Register Banks ***
                # sp, lr, and spsr are actually aliases to one of these.
                # Which one they reference depends on execution mode.
                # Emulators that care should be careful when loading state.
                # NOTE: Use User-mode copies of registers unless the mode has its own.
                # User-mode Stack Pointer
                "sp_usr": "sp_usr",
                # User-mode Link Register
                "lr_usr": "lr_usr",
                # User-mode r8
                "r8_usr": "r8_usr",
                # User-mode r9
                "r9_usr": "r9_usr",
                # User-mode r10
                "r10_usr": "r10_usr",
                # User-mode r11
                "r11_usr": "r11_usr",
                # User-mode r12
                "r12_usr": "r12_usr",
                # Hypervisor Stack Pointer
                "sp_hyp": "sp_hyp",
                # Hypervisor Saved PSR
                "spsr_hyp": "spsr_hyp",
                # Hypervisor Exception Link Register
                # NOTE: This isn't so much banked, as it only exists in hypervisor mode.
                "elr_hyp": "elr_hyp",
                # Supervisor Stack Pointer
                "sp_svc": "sp_svc",
                # Supervisor Link Register
                "lr_svc": "lr_svc",
                # Supervisor Saved PSR
                "spsr_svc": "spsr_svc",
                # Abort-state Stack Pointer
                "sp_abt": "sp_abt",
                # Abort-state Link Register
                "lr_abt": "lr_abt",
                # Abort-state Saved PSR
                "spsr_abt": "spsr_abt",
                # Undefined-mode Stack Pointer
                "sp_und": "sp_und",
                # Undefined-mode Link Register
                "lr_und": "lr_und",
                # Undefined-mode Saved PSR
                "spsr_und": "spsr_und",
                # Monitor-mode Stack Pointer
                "sp_mon": "sp_mon",
                # Monitor-mode Link Register
                "lr_mon": "lr_mon",
                # Monitor-mode Saved PSR
                "spsr_mon": "spsr_mon",
                # IRQ-mode Stack Pointer
                "sp_irq": "sp_irq",
                # IRQ-mode Link Register
                "lr_irq": "lr_irq",
                # IRQ-mode Saved PSR
                "spsr_irq": "spsr_irq",
                # FIQ-mode Stack Pointer
                "sp_fiq": "sp_fiq",
                # FIQ-mode Link Register
                "lr_fiq": "lr_fiq",
                # FIQ-mode Saved PSR
                "spsr_fiq": "spsr_fiq",
                # FIQ-mode r8
                "r8_fiq": "r8_fiq",
                # FIQ-mode r9
                "r9_fiq": "r9_fiq",
                # FIQ-mode r10
                "r10_fiq": "r10_fiq",
                # FIQ-mode r11
                "r11_fiq": "r11_fiq",
                # FIQ-mode r12
                "r12_fiq": "r12_fiq",
            }
        )


class ARMMachineMixinFP:
    """Mixin for ARM CPUs with FP extensions

    This is one kind of floating-point extension
    which offers 64-bit scalar operations.
    """

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Floating point control registers ***
                # Floating-point Status and Control Register
                "fpscr": "fpscr",
                # Floating-point Exception Control Register
                "fpexc": "fpexc",
                # Floating-point System ID Register
                "fpsid": "fpsid",
                # Media and VFP Feature Register 0
                "mvfr0": "mvfr0",
                # Media and VFP Feature Register 1
                "mvfr1": "mvfr1",
                # *** Floating point registers ***
                "d0": "d0",
                "s0": "s0",
                "s1": "s1",
                "d1": "d1",
                "s2": "s2",
                "s3": "s3",
                "d2": "d2",
                "s4": "s4",
                "s5": "s5",
                "d3": "d3",
                "s6": "s6",
                "s7": "s7",
                "d4": "d4",
                "s8": "s8",
                "s9": "s9",
                "d5": "d5",
                "s10": "s10",
                "s11": "s11",
                "d6": "d6",
                "s12": "s12",
                "s13": "s13",
                "d7": "d7",
                "s14": "s14",
                "s15": "s15",
                "d8": "d8",
                "s16": "s16",
                "s17": "s17",
                "d9": "d9",
                "s18": "s18",
                "s19": "s19",
                "d10": "d10",
                "s20": "s20",
                "s21": "s21",
                "d11": "d11",
                "s22": "s22",
                "s23": "s23",
                "d12": "d12",
                "s24": "s24",
                "s25": "s25",
                "d13": "d13",
                "s26": "s26",
                "s27": "s27",
                "d14": "d14",
                "s28": "s28",
                "s29": "s29",
                "d15": "d15",
                "s30": "s30",
                "s31": "s31",
            }
        )


class ARMMachineMixinVFPEL:
    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Floating-point Control Registers ***
                # Floating-point Status and Control Register
                "fpscr": "fpscr",
                # Floating-point Exception Control Register
                "fpexc": "fpexc",
                # Floating-point System ID Register
                "fpsid": "fpsid",
                # Media and VFP Feature Register 0
                "mvfr0": "mvfr0",
                # Media and VFP Feature Register 1
                "mvfr1": "mvfr1",
                # *** Floating-point Registers ****
                "q0": "q0",
                "d0": "d0",
                "s0": "s0",
                "s1": "s1",
                "d1": "d1",
                "s2": "s2",
                "s3": "s3",
                "q1": "q1",
                "d2": "d2",
                "s4": "s4",
                "s5": "s5",
                "d3": "d3",
                "s6": "s6",
                "s7": "s7",
                "q2": "q2",
                "d4": "d4",
                "s8": "s8",
                "s9": "s9",
                "d5": "d5",
                "s10": "s10",
                "s11": "s11",
                "q3": "q3",
                "d6": "d6",
                "s12": "s12",
                "s13": "s13",
                "d7": "d7",
                "s14": "s14",
                "s15": "s15",
                "q4": "q4",
                "d8": "d8",
                "s16": "s16",
                "s17": "s17",
                "d9": "d9",
                "s18": "s18",
                "s19": "s19",
                "q5": "q5",
                "d10": "d10",
                "s20": "s20",
                "s21": "s21",
                "d11": "d11",
                "s22": "s22",
                "s23": "s23",
                "q6": "q6",
                "d12": "d12",
                "s24": "s24",
                "s25": "s25",
                "d13": "d13",
                "s26": "s26",
                "s27": "s27",
                "q7": "q7",
                "d14": "d14",
                "s28": "s28",
                "s29": "s29",
                "d15": "d15",
                "s30": "s30",
                "s31": "s31",
                # NOTE: This isn't a typo; there are only 32 single-precision sX registers
                # This does mean that only half the VFP register space can be used
                # for single-precision arithmetic.
                "q8": "q8",
                "d16": "d16",
                "d17": "d17",
                "q9": "q9",
                "d18": "d18",
                "d19": "d19",
                "q10": "q10",
                "d20": "d20",
                "d21": "d21",
                "q11": "q11",
                "d22": "d22",
                "d23": "d23",
                "q12": "q12",
                "d24": "d24",
                "d25": "d25",
                "q13": "q13",
                "d26": "d26",
                "d27": "d27",
                "q14": "q14",
                "d28": "d28",
                "d29": "d29",
                "q15": "q15",
                "d30": "d30",
                "d31": "d31",
            }
        )


class ARMv5TMachineDef(ARMMachineMixinM, ARMMachineDef):
    angr_arch = archinfo.arch_arm.ArchARMEL()
    arch = Architecture.ARM_V5T
    byteorder = Byteorder.LITTLE


class ARMv6MMachineDef(ARMMachineMixinFP, ARMMachineMixinM, ARMMachineDef):
    angr_arch = archinfo.arch_arm.ArchARMEL()
    arch = Architecture.ARM_V6M
    byteorder = Byteorder.LITTLE


class ARMv6MThumbMachineDef(ARMv6MMachineDef):
    angr_arch = archinfo.arch_arm.ArchARMCortexM()
    arch = Architecture.ARM_V6M_THUMB
    byteorder = Byteorder.LITTLE
    is_thumb = True


class ARMv7MMachineDef(ARMMachineMixinFP, ARMMachineMixinM, ARMMachineDef):
    angr_arch = archinfo.arch_arm.ArchARMHF()
    arch = Architecture.ARM_V7M
    byteorder = Byteorder.LITTLE


class ARMv7AMachineDef(ARMMachineMixinVFPEL, ARMMachineMixinRA, ARMMachineDef):
    # TODO: This is a user-space integer only model
    angr_arch = archinfo.arch_arm.ArchARMHF()
    arch = Architecture.ARM_V7A
    byteorder = Byteorder.LITTLE


# angr does not have good enough R- or A- series support,
# or any support for NEON that I can see.



import copy

import angr

from ...utils import RangeCollection


class ExpandedScratchPlugin(angr.state_plugins.SimStateScratch):
    def __init__(self, scratch=None):
        super().__init__(scratch=scratch)
        self.exit_points = set()
        self.bounds = RangeCollection()
        self.memory_map = RangeCollection()
        self.global_insn_bp = None
        self.global_syscall_func = None
        self.global_read_bp = None
        self.global_write_bp = None
        self.insn_bps = dict()
        self.func_bps = dict()
        self.syscall_funcs = dict()
        self.mem_read_bps = dict()
        self.mem_write_bps = dict()
        self.extensions = dict()

        if scratch is not None:
            self.exit_points |= scratch.exit_points
            self.bounds.update(scratch.bounds)
            self.memory_map.update(scratch.memory_map)
            self.global_insn_bp = scratch.global_insn_bp
            self.global_syscall_func = scratch.global_syscall_func
            self.global_read_bp = scratch.global_read_bp
            self.global_write_bp = scratch.global_write_bp
            self.insn_bps.update(scratch.insn_bps)
            self.func_bps.update(scratch.func_bps)
            self.syscall_funcs.update(scratch.syscall_funcs)
            self.mem_read_bps.update(scratch.mem_read_bps)
            self.mem_write_bps.update(scratch.mem_write_bps)
            for name, ext in scratch.extensions.items():
                self.extensions[name] = copy.deepcopy(ext)

    @angr.SimStatePlugin.memo
    def copy(self, memo):
        return self.__class__(scratch=self)



import ctypes
import logging

import angr
import claripy

log = logging.getLogger(__name__)


class TypeDefPlugin(angr.SimStatePlugin):
    """Angr state plugin for tracking typedefs

    This captures two concepts for a state:

    - Type bindings for addresses, registers and symbols
    - Instance addresses for symbols
    """

    def __init__(self):
        super().__init__()
        self._structdefs = dict()

        self._symbols = dict()
        self._sym_values = dict()
        self._registers = dict()
        self._addrs = dict()
        self._allocations = dict()

    @angr.SimStatePlugin.memo
    def copy(self, memo):
        dup = TypeDefPlugin()
        dup._structdefs.update(self._structdefs)
        dup._symbols.update(self._symbols)
        dup._sym_values.update(self._sym_values)
        dup._registers.update(self._registers)

        # heckin' multi-layer copy...
        for typedef, addrs in self._allocations.items():
            dup._allocations[typedef] = set(addrs)

        return dup

    def ctype_to_bv(self, typedef, prefix="sw"):
        if hasattr(typedef, "_fields_"):
            off = 0
            fielddefs = list()
            for name, fielddef in typedef._fields_:
                field_off = getattr(typedef, name).offset
                if field_off > off:
                    # handle padding
                    fielddefs.append(claripy.BVS("padding", (field_off - off) * 8))
                    off = field_off
                fielddefs.append(self.ctype_to_bv(fielddef, f"{prefix}_{name}"))
                off += ctypes.sizeof(fielddef)

            return claripy.Concat(*fielddefs)
        else:
            res = claripy.BVS(prefix, ctypes.sizeof(typedef) * 8)
            self.bind_symbol(res, typedef)
            res = claripy.Reverse(res)
            return res

    def add_struct(self, structdef):
        """Add a struct type to this context for easy lookup"""
        if structdef._name in self._structdefs:
            raise ValueError(f"Already have a struct def for {structdef._name}")
        self._structdefs[structdef._name] = structdef

    def get_struct(self, name):
        """Look up a struct type by its name"""
        return self._structdefs[name]

    def bind_symbol(self, sym, typedef):
        """Bind a symbol to a typedef"""
        if sym in self._symbols:
            raise ValueError(f"Already have a binding for {sym}")
        self._symbols[sym.args[0]] = typedef

    def get_symbol_binding(self, sym):
        """Get the type binding for a symbol"""
        if sym in self._symbols:
            return self._symbols[sym]
        return None

    def set_symbol(self, sym, val):
        """Set a concrete value for a symbol

        Symbols are static-single-assignment,
        so if they are collapsed to a concrete value,
        that assignment will stick.
        The assignment is normally bound up in the solver,
        making it annoying and expensive to recover.
        This feature lets us quickly recover our stuff
        """
        log.info(f"Setting {sym} to {val}")
        if sym in self._sym_values:
            raise ValueError(f"Already have value {self._sym_values[sym]} for {sym}")
        self._sym_values[sym] = val

    def get_symbol(self, sym):
        """Get a concrete value for a symbol"""
        if sym in self._sym_values:
            return self._sym_values[sym]
        return None

    def bind_register(self, name, typedef):
        """Bind a register to a typedef

        Registers themselves don't have specific types.
        However, this only gets used in uninitialized value analysis;
        it only covers the type a register has before it's
        written by the program.
        """
        if name in self._registers:
            raise ValueError(f"Already have a binding for {name}")
        self._registers[name] = typedef

    def get_register_binding(self, name):
        """Get the type binding for a register

        Registers themselves don't have specific types.
        However, this only gets used in uninitialized value analysis;
        it only covers the type a register has before it's
        written by the program.
        """
        if name in self._registers:
            return self._registers[name]
        return None

    def bind_address(self, addr, typedef):
        if addr in self._addrs:
            raise ValueError(f"Already have a binding for {addr}")
        self._addrs[addr] = typedef

    def get_address_binding(self, addr):
        if addr in self._addrs:
            return self._addrs[addr]
        return None

    def allocate(self, typedef):
        addr = self.state.heap.allocate(ctypes.sizeof(typedef))

        rep = self.ctype_to_bv(typedef)

        self._allocations.setdefault(typedef, set()).add(addr)
        log.debug(f"Storing {rep} at {addr:x}")
        self.state.memory.store(addr, rep)
        return addr

    def get_allocs_for_type(self, typedef):
        if typedef in self._allocations:
            return list(self._allocations[typedef])
        return list()

    def pp(self, log, all_allocs=True):
        log("Register Bindings")
        for reg_name, reg_def in self._registers.items():
            log(f"\t{reg_name} -> {reg_def}")
        log("Symbol Bindings")
        for sym_name, sym_def in self._symbols.items():
            if sym_name in self._sym_values:
                log(f"\t{sym_name} -> {sym_def} ({self._sym_values[sym_name]:x})")
            else:
                log(f"\t{sym_name} -> {sym_def} (No Value)")
        if all_allocs:
            log("Allocations:")
            for typedef, addrs in self._allocations.items():
                log(f"- {typedef}")
                for addr in addrs:
                    log(f"\t- {addr:x}")



import angr


def print_state(log, state, tag):
    """Pretty-print data contained in an execution state

    This prints useful registers, memory, and state constraints.
    """
    log(f"{state}: {tag}")
    if not state.regs.rip.symbolic:
        try:
            block = state.block()
            block.pp()
        except angr.errors.SimEngineError:
            log("(Invalid IP!)")
    else:
        log("(Exit state)")
    log("Registers")
    state.registers.pp(log)
    log("Memory")
    state.memory.pp(log)
    log("Constraints:")
    for expr in state.solver.constraints:
        log(f"\t{expr}")
    state.typedefs.pp(log)



from .analysis import *  # noqa: F401, F403
from .analysis import __all__ as __analysis__
from .colorizer import Colorizer
from .colorizer_def_use import ColorizerDefUse
from .colorizer_summary import ColorizerSummary
from .coverage_frontier import CoverageFrontier
from .field_detection import FieldDetectionAnalysis, ForcedFieldDetectionAnalysis
from .forced_exec import ForcedExecution
from .trace_execution import TraceExecution
from .trace_execution_types import CmpInfo, TraceElement, TraceRes

__all__ = __analysis__ + [
    "Colorizer",
    "ColorizerSummary",
    "ColorizerDefUse",
    "FieldDetectionAnalysis",
    "ForcedFieldDetectionAnalysis",
    "ForcedExecution",
    "TraceExecution",
    "TraceElement",
    "TraceRes",
    "CmpInfo",
    "CoverageFrontier",
]



import typing

from .... import platforms
from . import stack


class MIPSStack(stack.DescendingStack):
    """A stack for a MIPS 32-bit CPU"""

    def get_alignment(self) -> int:
        return 4

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for MIPS32")


class MIPSBEStack(MIPSStack):
    """A stack for a big-endian MIPS 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.MIPS32, platforms.Byteorder.BIG
    )


class MIPSELStack(MIPSStack):
    """A stack for a little-endian MIPS 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.MIPS32, platforms.Byteorder.LITTLE
    )



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import MIPSSysVModel


class MIPSSysVBsdSignal(BsdSignal, MIPSSysVModel):
    pass


class MIPSSysVKill(Kill, MIPSSysVModel):
    pass


class MIPSSysVKillpg(Killpg, MIPSSysVModel):
    pass


class MIPSSysVPthreadKill(PthreadKill, MIPSSysVModel):
    pass


class MIPSSysVPthreadSigmask(PthreadSigmask, MIPSSysVModel):
    pass


class MIPSSysVSigaction(Sigaction, MIPSSysVModel):
    pass


class MIPSSysVSigaddset(Sigaddset, MIPSSysVModel):
    pass


class MIPSSysVSigaltstack(Sigaltstack, MIPSSysVModel):
    pass


class MIPSSysVSigdelset(Sigdelset, MIPSSysVModel):
    pass


class MIPSSysVSigemptyset(Sigemptyset, MIPSSysVModel):
    pass


class MIPSSysVSigfillset(Sigfillset, MIPSSysVModel):
    pass


class MIPSSysVSighold(Sighold, MIPSSysVModel):
    pass


class MIPSSysVSigignore(Sigignore, MIPSSysVModel):
    pass


class MIPSSysVSiginterrupt(Siginterrupt, MIPSSysVModel):
    pass


class MIPSSysVSigismember(Sigismember, MIPSSysVModel):
    pass


class MIPSSysVSigpause(Sigpause, MIPSSysVModel):
    pass


class MIPSSysVSigpending(Sigpending, MIPSSysVModel):
    pass


class MIPSSysVSigprocmask(Sigprocmask, MIPSSysVModel):
    pass


class MIPSSysVSigqueue(Sigqueue, MIPSSysVModel):
    pass


class MIPSSysVSigrelse(Sigrelse, MIPSSysVModel):
    pass


class MIPSSysVSigset(Sigset, MIPSSysVModel):
    pass


class MIPSSysVSigsuspend(Sigsuspend, MIPSSysVModel):
    pass


class MIPSSysVSigtimedwait(Sigtimedwait, MIPSSysVModel):
    pass


class MIPSSysVSigwait(Sigwait, MIPSSysVModel):
    pass


class MIPSSysVSigwaitinfo(Sigwaitinfo, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVBsdSignal",
    "MIPSSysVKill",
    "MIPSSysVKillpg",
    "MIPSSysVPthreadKill",
    "MIPSSysVPthreadSigmask",
    "MIPSSysVSigaction",
    "MIPSSysVSigaddset",
    "MIPSSysVSigaltstack",
    "MIPSSysVSigdelset",
    "MIPSSysVSigemptyset",
    "MIPSSysVSigfillset",
    "MIPSSysVSighold",
    "MIPSSysVSigignore",
    "MIPSSysVSiginterrupt",
    "MIPSSysVSigismember",
    "MIPSSysVSigpause",
    "MIPSSysVSigpending",
    "MIPSSysVSigprocmask",
    "MIPSSysVSigqueue",
    "MIPSSysVSigrelse",
    "MIPSSysVSigset",
    "MIPSSysVSigsuspend",
    "MIPSSysVSigtimedwait",
    "MIPSSysVSigwait",
    "MIPSSysVSigwaitinfo",
]



import enum
import typing

import angr
import archinfo
import pypcode

from ....exceptions import EmulationError
from ....platforms import Architecture, Byteorder
from .machdef import GhidraMachineDef


class UpdatedEnumMeta(enum.EnumMeta):
    def __contains__(cls, obj):
        if isinstance(obj, int):
            return obj in cls._value2member_map_
        return enum.EnumMeta.__contains__(enum.EnumMeta, obj)


def handle_nop(irsb, i):
    # This op has no impact on user-facing machine state.
    irsb._ops.pop(i)
    return i


def handle_sigtrap(irsb, i):
    # This op should terminate this block with SIGTRAP
    next_addr = irsb._ops[i - 1].inputs[0].offset + 3
    irsb._ops = irsb._ops[0:i]
    irsb.next = next_addr
    irsb._size = next_addr - irsb.addr
    irsb._instruction_addresses = list(
        filter(lambda x: x < next_addr, irsb._instruction_addresses)
    )
    irsb.jumpkind = "Ijk_SigTRAP"
    return i


def handle_syscall(irsb, i):
    # This op should terminate this block with a syscall
    next_addr = irsb._ops[i - 1].inputs[0].offset + 3
    irsb._ops = irsb._ops[0:i]
    irsb.next = next_addr
    irsb._size = next_addr - irsb.addr
    irsb._instruction_addresses = list(
        filter(lambda x: x < next_addr, irsb._instruction_addresses)
    )
    irsb.jumpkind = "Ijk_Sys_syscall"

    return i


def handle_unimpl(irsb, i):
    # I don't know what this op does yet
    userop = LoongArchUserOp(irsb._ops[i].inputs[0].offset)
    raise EmulationError(f"Unimplemented user op {userop!r}")


class LoongArchUserOp(enum.IntEnum, metaclass=UpdatedEnumMeta):
    def __new__(
        cls, val: int, name: str = "", handler: typing.Any = None, desc: str = ""
    ):
        obj = int.__new__(cls, val)
        obj._value_ = val
        obj.short_name = name
        obj.handler = handler
        obj.description = desc
        return obj

    def __init__(
        self, val: int, name: str = "", handler: typing.Any = None, desc: str = ""
    ):
        self._value_: int = val
        self.short_name: str = name
        self.handler: typing.Any = handler
        self.description: str = desc

    def __repr__(self):
        return f"{hex(self.value)}: {self.short_name} - {self.description}"

    BREAK = 0x00, "break", handle_sigtrap, "Breakpoint instruction"
    CPUCFG = 0x01, "cpucfg", handle_unimpl, "CPU config instruction"
    ADDR_BOUND_EXCEPTION = 0x02, "addr_bound_exception", handle_unimpl, "Unknown"
    BOUND_CHECK_EXCEPTION = 0x03, "bound_check_exception", handle_unimpl, "Unknown"
    CRC_IEEE802_3 = 0x04, "crc_ieee802.3", handle_unimpl, "Unknown"
    CRC_CASTAGNOLI = 0x05, "crc_castagnoli", handle_unimpl, "Unknown"
    DBCL = 0x06, "dbcl", handle_unimpl, "Unknown"
    DBAR = 0x07, "dbar", handle_nop, "Data Barrier"
    IBAR = 0x08, "ibar", handle_nop, "Instruction Barrier"
    IOCSRRD = 0x09, "iocsrrd", handle_unimpl, "Unknown"
    IOCSRWR = 0x0A, "iocsrwr", handle_unimpl, "Unknown"
    PRELD_LOADL1CACHE = 0x0B, "preld_loadl1cache", handle_unimpl, "Unknown"
    PRELD_STOREL1CACHE = 0x0C, "preld_storel1cache", handle_unimpl, "Unknown"
    PRELD_NOP = 0x0D, "preld_nop", handle_unimpl, "Unknown"
    PRELDX_LOADL1CACHE = 0x0E, "preldx_loadl1cache", handle_unimpl, "Unknown"
    PRELDX_STOREL1CACHE = 0x0F, "preldx_storel1cache", handle_unimpl, "Unknown"
    PRELDX_NOP = 0x10, "preldx_nop", handle_unimpl, "Unknown"
    RDTIME_COUNTER = 0x11, "rdtime.counter", handle_unimpl, "Unknown"
    RDTIME_COUNTERID = 0x12, "rdtime.counterid", handle_unimpl, "Unknown"
    SYSCALL = 0x13, "syscall", handle_syscall, "Unknown"
    F_SCALEB = 0x14, "f_scaleb", handle_unimpl, "Unknown"
    F_LOGB = 0x15, "f_logb", handle_unimpl, "Unknown"
    F_CLASS = 0x16, "f_class", handle_unimpl, "Unknown"
    ROUND_EVEN = 0x17, "round_even", handle_unimpl, "Unknown"


class LoongArchMachineDef(GhidraMachineDef):
    byteorder = Byteorder.LITTLE

    _registers = {
        "pc": "pc",
        # Zero register
        "r0": "zero",
        "zero": "zero",
        # Return address
        "r1": "ra",
        "ra": "ra",
        # TLS pointer
        "r2": "tp",
        "tp": "tp",
        # Stack pointer
        "r3": "sp",
        "sp": "sp",
        # Arguments.
        # a0 and a1 are also the return registers
        "r4": "a0",
        "a0": "a0",
        "v0": "a0",
        "r5": "a1",
        "a1": "a1",
        "v1": "a1",
        "r6": "a2",
        "a2": "a2",
        "r7": "a3",
        "a3": "a3",
        "r8": "a4",
        "a4": "a4",
        "r9": "a5",
        "a5": "a5",
        "r10": "a6",
        "a6": "a6",
        "r11": "a7",
        "a7": "a7",
        # Temporary registers
        "r12": "t0",
        "t0": "t0",
        "r13": "t1",
        "t1": "t1",
        "r14": "t2",
        "t2": "t2",
        "r15": "t3",
        "t3": "t3",
        "r16": "t4",
        "t4": "t4",
        "r17": "t5",
        "t5": "t5",
        "r18": "t6",
        "t6": "t6",
        "r19": "t7",
        "t7": "t7",
        "r20": "t8",
        "t8": "t8",
        # Per-CPU Base Address
        "r21": "r21",
        "u0": "r21",
        # Frame Pointer
        "r22": "fp",
        "fp": "fp",
        # Static registers
        "r23": "s0",
        "s0": "s0",
        "r24": "s1",
        "s1": "s1",
        "r25": "s2",
        "s2": "s2",
        "r26": "s3",
        "s3": "s3",
        "r27": "s4",
        "s4": "s4",
        "r28": "s5",
        "s5": "s5",
        "r29": "s6",
        "s6": "s6",
        "r30": "s7",
        "s7": "s7",
        "r31": "s8",
        "s8": "s8",
        # Floating-point arguments.
        # fa0 and fa1 are also return values
        "f0": "fa0",
        "fa0": "fa0",
        "f1": "fa1",
        "fa1": "fa1",
        "f2": "fa2",
        "fa2": "fa2",
        "f3": "fa3",
        "fa3": "fa3",
        "f4": "fa4",
        "fa4": "fa4",
        "f5": "fa5",
        "fa5": "fa5",
        "f6": "fa6",
        "fa6": "fa6",
        "f7": "fa7",
        "fa7": "fa7",
        # Floating-point temporary registers
        "f8": "ft0",
        "ft0": "ft0",
        "f9": "ft1",
        "ft1": "ft1",
        "f10": "ft2",
        "ft2": "ft2",
        "f11": "ft3",
        "ft3": "ft3",
        "f12": "ft4",
        "ft4": "ft4",
        "f13": "ft5",
        "ft5": "ft5",
        "f14": "ft6",
        "ft6": "ft6",
        "f15": "ft7",
        "ft7": "ft7",
        "f16": "ft8",
        "ft8": "ft8",
        "f17": "ft9",
        "ft9": "ft9",
        "f18": "ft10",
        "ft10": "ft10",
        "f19": "ft11",
        "ft11": "ft11",
        "f20": "ft12",
        "ft12": "ft12",
        "f21": "ft13",
        "ft13": "ft13",
        "f22": "ft14",
        "ft14": "ft14",
        "f23": "ft15",
        "ft15": "ft15",
        # Floating-point static registers
        "f24": "fs0",
        "fs0": "fs0",
        "f25": "fs1",
        "fs1": "fs1",
        "f26": "fs2",
        "fs2": "fs2",
        "f27": "fs3",
        "fs3": "fs3",
        "f28": "fs4",
        "fs4": "fs4",
        "f29": "fs5",
        "fs5": "fs5",
        "f30": "fs6",
        "fs6": "fs6",
        "f31": "fs7",
        "fs7": "fs7",
    }

    def successors(self, state: angr.SimState, **kwargs) -> typing.Any:
        # xtensa includes a _LOT_ of custom pcode operations.

        # Fetch or compute the IR block for our state
        if "irsb" in kwargs and kwargs["irsb"] is not None:
            # Someone's already specified an IR block.
            irsb = kwargs["irsb"]
        else:
            # Disable optimization; it doesn't work
            kwargs["opt_level"] = 0

            # Compute the block from the state.
            # Pray to the Powers that kwargs are compatible.
            irsb = state.block(**kwargs).vex

        i = 0
        while i < len(irsb._ops):
            op = irsb._ops[i]
            if op.opcode == pypcode.OpCode.CALLOTHER:
                # This is a user-defined Pcode op.
                # Alter irsb to mimic its behavior, if we can.
                opnum = op.inputs[0].offset

                if opnum not in LoongArchUserOp:
                    # Not a userop.
                    raise EmulationError(f"Undefined user op {hex(opnum)}")
                # get the enum struct
                userop = LoongArchUserOp(opnum)

                # Invoke the handler
                i = userop.handler(irsb, i)
            else:
                i += 1

        # Force the engine to use our IR block
        kwargs["irsb"] = irsb

        # Turn the crank on the engine
        return super().successors(state, **kwargs)


class SimCCLoongArchLinux(angr.calling_conventions.SimCC):
    ARG_REGS = ["a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"]
    FP_ARG_REGS = ["fa0", "fa1", "fa2", "fa3", "fa4", "fa5", "fa6", "fa7"]
    RETURN_VAL = angr.calling_conventions.SimRegArg("a0", 8)
    RETURN_ADDR = angr.calling_conventions.SimRegArg("ra", 8)
    ARCH = archinfo.ArchPcode("Loongarch:LE:64:lp64d")  # type: ignore


angr.calling_conventions.register_default_cc(
    "Loongarch:LE:64:lp64d", SimCCLoongArchLinux
)


class SimCCLoongArchLinuxSyscall(angr.calling_conventions.SimCCSyscall):
    ARG_REGS = ["a0", "a1", "a2", "a3", "a4", "a5"]
    FP_ARG_REGS: typing.List[str] = []
    RETURN_VAL = angr.calling_conventions.SimRegArg("a0", 8)
    RETURN_ADDR = angr.calling_conventions.SimRegArg("ip_at_syscall", 8)
    ARCH = None

    @classmethod
    def _match(cls, arch, args, sp_data):
        # Never match; only occurs durring syscalls
        return False

    @staticmethod
    def syscall_num(state):
        # This is a massive guess from RE'ing libc
        return state.regs.a7


angr.calling_conventions.register_syscall_cc(
    "Loongarch:LE:64:lp64d", "default", SimCCLoongArchLinuxSyscall
)


class LoongArch64MachineDef(LoongArchMachineDef):
    arch = Architecture.LOONGARCH64
    pcode_language = "Loongarch:LE:64:lp64d"


__all__ = ["LoongArch64MachineDef"]



import abc
import enum
import logging
import struct
import typing

from ... import emulators, exceptions
from ...platforms import Byteorder, PlatformDef
from .model import Model

log = logging.getLogger(__name__)


class ArgumentType(enum.Enum):
    """C primitive data types for specifying function arguments.

    The exact size of these types depends strongly on the ABI.
    These specify the source-level signature;
    the ABI-specific subclasses of CStdModel figure out how to decode them.
    """

    CHAR = "char"
    UCHAR = "unsigned char"

    SHORT = "short"
    USHORT = "unsigned short"

    INT = "int"
    UINT = "unsigned int"

    LONG = "long"
    ULONG = "unsigned long"

    POINTER = "pointer"

    SIZE_T = "size_t"
    SSIZE_T = "ssize_t"

    LONGLONG = "long long"
    ULONGLONG = "unsigned long long"

    FLOAT = "float"
    DOUBLE = "double"

    VOID = "void"


signed_int_types = {
    ArgumentType.INT,
    ArgumentType.LONG,
    ArgumentType.LONGLONG,
    ArgumentType.SSIZE_T,
}


def add_argument(
    i: int, kind: ArgumentType, model: typing.Union["CStdModel", "VariadicContext"]
):
    if kind in model._four_byte_types or (
        kind == ArgumentType.FLOAT and model._soft_float
    ):
        # Four byte int type, or float on a system without separate FP arg regs
        if model._int_reg_offset == len(model._four_byte_arg_regs):
            # No room left in registers; use stack
            model._on_stack.append(True)
            model._arg_offset.append(model._stack_offset + model._init_stack_offset)
            model._stack_offset += model._four_byte_stack_size
        else:
            # Registers left; use them
            model._on_stack.append(False)
            model._arg_offset.append(model._int_reg_offset)
            model._int_reg_offset += 1
    elif kind in model._eight_byte_types or (
        kind == ArgumentType.DOUBLE and model._soft_float
    ):
        # Eight byte int type, or double on a system without separate FP arg regs
        if model._int_reg_offset % model._eight_byte_reg_size != 0:
            # Align argument register for eight-byte value
            model._int_reg_offset += 1

        if model._int_reg_offset == len(model._eight_byte_arg_regs):
            # No room left in registers; use stack
            if (
                model._align_stack
                and model._stack_offset % model._eight_byte_stack_size != 0
            ):
                # Stack out of alignment.  Align it.
                model._stack_offset += 4
            model._on_stack.append(True)
            model._arg_offset.append(model._stack_offset + model._init_stack_offset)
            model._stack_offset += model._eight_byte_stack_size
        else:
            # Registers left; use them
            model._on_stack.append(False)
            model._arg_offset.append(model._int_reg_offset)
            model._int_reg_offset += model._eight_byte_reg_size
    elif kind == ArgumentType.FLOAT:
        # Float type
        if model._fp_reg_offset == len(model._float_arg_regs):
            # No room left in registers; use stack
            model._on_stack.append(True)
            model._arg_offset.append(model._stack_offset + model._init_stack_offset)
            model._stack_offset += model._float_stack_size
        else:
            # Registers left; use them
            model._on_stack.append(False)
            model._arg_offset.append(model._fp_reg_offset)
            model._fp_reg_offset += 1
    elif kind == ArgumentType.DOUBLE:
        # Double type
        if model._fp_reg_offset % model._double_reg_size != 0:
            model._fp_reg_offset += 1

        if model._fp_reg_offset == len(model._double_arg_regs):
            # No room left in registers; use stack
            if (
                model._align_stack
                and model._stack_offset % model._double_stack_size != 0
            ):
                model._stack_offset += 4
            model._on_stack.append(True)
            model._arg_offset.append(model._stack_offset + model._init_stack_offset)
            model._stack_offset += model._double_stack_size
        else:
            # Registers left; use them
            model._on_stack.append(False)
            model._arg_offset.append(model._fp_reg_offset)
            model._fp_reg_offset += model._double_reg_size
    else:
        raise exceptions.ConfigurationError(
            f"{model.name} argument {i} has unknown type {kind}"
        )


def get_argument(
    model: typing.Union["CStdModel", "VariadicContext"],
    index: int,
    kind: ArgumentType,
    emulator: emulators.Emulator,
) -> typing.Union[int, float]:
    # Get an argument out of a CStdModel or VariadicContext
    sp = model.platdef.sp_register
    on_stack = model._on_stack[index]
    arg_offset = model._arg_offset[index]

    if kind in model._four_byte_types:
        # Four byte integer
        if on_stack:
            # Stored on the stack; read from memory
            addr = emulator.read_register(sp) + arg_offset
            data = emulator.read_memory(addr, model._four_byte_stack_size)
            if model.platform.byteorder == Byteorder.BIG:
                intval = int.from_bytes(data, "big")
            else:
                intval = int.from_bytes(data, "little")
        else:
            # Stored in a register
            intval = emulator.read_register(model._four_byte_arg_regs[arg_offset])

        # Handle integer signing and signedness
        # SmallWorld registers are unsigned, so we'll need to convert.
        # Some architectures zero-extend their integers, so we need to mask.
        intval = intval & model._int_inv_mask
        if kind in signed_int_types and (intval & model._int_sign_mask) != 0:
            intval = (intval ^ model._int_inv_mask) + 1
            intval *= -1
        return intval

    elif kind in model._eight_byte_types:
        # Eight byte integer
        if on_stack:
            # Stored on the stack
            addr = emulator.read_register(sp) + arg_offset
            data = emulator.read_memory(addr, model._eight_byte_stack_size)
            if model.platform.byteorder == Byteorder.BIG:
                intval = int.from_bytes(data, "big")
            else:
                intval = int.from_bytes(data, "little")
        elif model._eight_byte_reg_size == 2:
            # Stored in a register pair
            lo = emulator.read_register(model._eight_byte_arg_regs[arg_offset])
            hi = emulator.read_register(model._eight_byte_arg_regs[arg_offset + 1])
            if model.platform.byteorder == Byteorder.BIG:
                tmp = lo
                lo = hi
                hi = tmp
            intval = (hi << 32) | lo
        else:
            # Stored in a single register
            intval = emulator.read_register(model._eight_byte_arg_regs[arg_offset])

        # Handle signedness
        # SmallWorld registers are unsigned, so we'll need to convert.
        if kind in signed_int_types and (intval & model._long_long_sign_mask) != 0:
            intval = (intval ^ model._long_long_inv_mask) + 1
            intval *= -1
        return intval
    elif kind == ArgumentType.FLOAT:
        # Four-byte float
        if on_stack:
            # Stored on the stack
            addr = emulator.read_register(sp) + arg_offset
            data = emulator.read_memory(addr, model._float_stack_size)
            if model.platform.byteorder == Byteorder.BIG:
                intval = int.from_bytes(data, "big")
            else:
                intval = int.from_bytes(data, "little")
        elif model._soft_float:
            # Soft-float ABI; treat as a four-byte int
            intval = emulator.read_register(model._four_byte_arg_regs[arg_offset])
        else:
            # Hard-float ABI; fetch from FPU registers
            intval = emulator.read_register(model._float_arg_regs[arg_offset])

        # Unpack the bits into a Python float
        # SmallWorld already did the work of converting endianness.
        if model._floats_are_doubles:
            # Some ABIs promote floats to doubles.
            # And by "some ABIs", I mean PowerPC.
            byteval = intval.to_bytes(8, "little")
            (floatval,) = struct.unpack("<d", byteval)
        else:
            byteval = (intval & model._int_inv_mask).to_bytes(4, "little")
            (floatval,) = struct.unpack("<f", byteval)
        return floatval
    elif kind == ArgumentType.DOUBLE:
        # Eight-byte double float
        if on_stack:
            # Stored on the stack
            addr = emulator.read_register(sp) + arg_offset
            data = emulator.read_memory(addr, model._double_stack_size)
            if model.platform.byteorder == Byteorder.BIG:
                intval = int.from_bytes(data, "big")
            else:
                intval = int.from_bytes(data, "little")
        else:
            if model._soft_float:
                # Soft-float ABI; treated as an eight-byte int
                reg_array = model._eight_byte_arg_regs
                n_regs = model._eight_byte_reg_size
            else:
                # Hard-float ABI; stored in FPU registers
                reg_array = model._double_arg_regs
                n_regs = model._double_reg_size

            if n_regs == 2:
                # Register pair.  Possible for both soft and hard floats.
                lo = emulator.read_register(reg_array[arg_offset])
                hi = emulator.read_register(reg_array[arg_offset + 1])
                if model.platform.byteorder == Byteorder.BIG:
                    tmp = lo
                    lo = hi
                    hi = tmp
                intval = (hi << 32) | lo
            else:
                # Single register
                intval = emulator.read_register(reg_array[arg_offset])

        # Convert bits into Python float
        # SmallWorld already did the work of converting endianness.
        byteval = intval.to_bytes(8, "little")
        (floatval,) = struct.unpack("<d", byteval)
        return floatval
    else:
        raise exceptions.ConfigurationError(
            f"{model.name} argument {index} has unknown type {kind}"
        )


class CStdModel(Model):
    """Base class for C standard function models


    Regardless of which version of a library you use,
    all "true" C functions will use the same interface
    defined by the ABI.
    (There are exceptions, such as thunks and internal functions
    never intended for human eyes)

    This abstracts away the ABI-specific operations
    performed by a function, namely getting args and returning vals.

    This also includes a parameterizable calling convention model.
    Every calling convention I've seen fits into a kind
    of Grand Unifying Theory.  They all pass a certain
    number of arguments via registers before switching to stack,
    and handle different-sized integers or floats
    in a few standard ways.

    It was way easier to figure out this theory and write one model
    than to maintain eleven-plus separate models.
    This may break for particularly unusual architectures,
    or very strange ABIs.  For living architectures
    with Debian support (sorry, hppa), it works.
    """

    # Flag indicating this model is imprecise.
    #
    # Most models are assumed to be approximations,
    # but this model definitely doesn't capture
    # a critical behavior.
    #
    # By default, these models should raise an exception if called.
    # The user can accept the risk and run a placeholde version
    # by setting the attribute "allow_imprecise" to True.
    #
    # Authors probably shouldn't rely on this flag
    # to mark truly-unimplemented models;
    # just raise an exception yourself.
    imprecise = False

    @property
    @abc.abstractmethod
    def argument_types(self) -> typing.List[ArgumentType]:
        """List of argument types for this function

        NOTE: Don't include variadics.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def return_type(self) -> ArgumentType:
        """Return type for this function"""
        raise NotImplementedError()

    def __init__(self, address: int):
        super().__init__(address)

        self.platdef = PlatformDef.for_platform(self.platform)

        # Set this to True to bypass the "imprecise" flag.
        self.allow_imprecise = False

        self._int_reg_offset = 0
        self._fp_reg_offset = 0
        self._stack_offset = 0

        self._on_stack: typing.List[bool] = list()
        self._arg_offset: typing.List[int] = list()

        for i in range(0, len(self.argument_types)):
            t = self.argument_types[i]
            add_argument(i, t, self)

    def model(self, emulator: emulators.Emulator):
        if self.imprecise and not self.allow_imprecise:
            raise exceptions.ConfigurationError(
                f"Invoked model for {self.name}, which is imprecise"
            )

    @abc.abstractmethod
    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return a four-byte type"""
        raise NotImplementedError()

    @abc.abstractmethod
    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return an eight-byte type"""
        raise NotImplementedError()

    @abc.abstractmethod
    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a float"""
        raise NotImplementedError()

    @abc.abstractmethod
    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a double"""
        raise NotImplementedError()

    def get_arg1(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the first argument from the emulator"""
        return get_argument(self, 0, self.argument_types[0], emulator)

    def get_arg2(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the second argument from the emulator"""
        return get_argument(self, 1, self.argument_types[1], emulator)

    def get_arg3(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the first argument from the emulator"""
        return get_argument(self, 2, self.argument_types[2], emulator)

    def get_arg4(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the first argument from the emulator"""
        return get_argument(self, 3, self.argument_types[3], emulator)

    def get_arg5(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the first argument from the emulator"""
        return get_argument(self, 4, self.argument_types[4], emulator)

    def get_arg6(self, emulator: emulators.Emulator) -> typing.Union[int, float]:
        """Fetch the first argument from the emulator"""
        return get_argument(self, 5, self.argument_types[5], emulator)

    def get_varargs(self) -> "VariadicContext":
        """Get a variadic argument context to fetch varargs

        Also necessary to handle more than six arguments.
        """
        return VariadicContext(self)

    def set_return_value(
        self, emulator: emulators.Emulator, val: typing.Union[int, float]
    ) -> None:
        """Return according to the appropriate return type"""
        if self.return_type == ArgumentType.VOID:
            # We're void.  You can't return from void.
            raise exceptions.ConfigurationError(
                f"{self.name} returning from void function"
            )

        if self.return_type == ArgumentType.FLOAT:
            # We're a float.
            if not isinstance(val, float):
                raise exceptions.ConfigurationError(
                    f"{self.name} trying to return {type(val)} as a float"
                )
            self._return_float(emulator, val)
            return

        if self.return_type == ArgumentType.DOUBLE:
            # We're a double
            if not isinstance(val, float):
                raise exceptions.ConfigurationError(
                    f"{self.name} trying to return {type(val)} as a double"
                )
            self._return_double(emulator, val)
            return

        # All other types are integral
        if not isinstance(val, int):
            raise exceptions.ConfigurationError(
                f"{self.name} trying to return {type(val)} as an integral type"
            )

        if val < 0:
            # Negative value; need to find 2s-compliment if it's an int
            val *= -1
            if self.return_type in self._four_byte_types:
                val = ((val ^ self._int_inv_mask) + 1) & self._int_inv_mask
            elif self.return_type in self._eight_byte_types:
                val = ((val ^ self._long_long_inv_mask) + 1) & self._long_long_inv_mask
            elif (
                self.return_type == ArgumentType.FLOAT
                or self.return_type == ArgumentType.DOUBLE
            ):
                # Floating-point type; encoding will take care of this.
                pass
            else:
                # Unsigned type; why are you returning a negative?
                raise exceptions.ConfigurationError(
                    f"{self.name} tried to return a signed value"
                )

        # Delegate return to handler
        if self.return_type in self._four_byte_types:
            self._return_4_byte(emulator, val)

        elif self.return_type in self._eight_byte_types:
            self._return_8_byte(emulator, val)

        else:
            raise exceptions.ConfigurationError(
                f"{self.name} returning unhandled type {self.return_type}"
            )

    def read_integer(
        self, address: int, kind: ArgumentType, emulator: emulators.Emulator
    ) -> int:
        """Read an integer out of memory based on type

        Arguments:
            address: The address to read from
            kind: The ArgumentType to read
            emulator: The emulator to read from
        """
        if kind in (ArgumentType.CHAR, ArgumentType.UCHAR):
            width = 1
        elif kind in (ArgumentType.SHORT, ArgumentType.USHORT):
            width = 2
        elif kind in (ArgumentType.INT, ArgumentType.UINT):
            width = 4
        elif kind in (ArgumentType.LONG, ArgumentType.ULONG, ArgumentType.POINTER):
            width = 4 if ArgumentType.LONG in self._four_byte_types else 8
        elif kind in (ArgumentType.LONGLONG, ArgumentType.ULONGLONG):
            width = 8

        byteval = emulator.read_memory(address, width)

        if self.platform.byteorder == Byteorder.LITTLE:
            return int.from_bytes(byteval, "little")
        else:
            return int.from_bytes(byteval, "big")

    def write_integer(
        self,
        address: int,
        intval: int,
        kind: ArgumentType,
        emulator: emulators.Emulator,
    ) -> None:
        """Write an integer based on type

        Note that this will handle converting signed to bytes.

        Arguments:
            address: The address to write to
            intval: The integer value to write
            kind: The ArgumentType to write
            emulator: The emulator to write to
        """
        if intval < 0:
            intval *= -1
            intval = (intval ^ self._long_long_inv_mask) + 1

        if kind in (ArgumentType.CHAR, ArgumentType.UCHAR):
            intval &= self._char_inv_mask
            width = 1
        elif kind in (ArgumentType.SHORT, ArgumentType.USHORT):
            intval &= self._short_inv_mask
            width = 2
        elif kind in (ArgumentType.INT, ArgumentType.UINT):
            intval &= self._int_inv_mask
            width = 4
        elif kind in (ArgumentType.LONG, ArgumentType.ULONG, ArgumentType.POINTER):
            intval &= self._long_inv_mask
            width = 4 if ArgumentType.LONG in self._four_byte_types else 8
        elif kind in (ArgumentType.LONGLONG, ArgumentType.ULONGLONG):
            intval &= self._long_long_inv_mask
            width = 8

        if self.platform.byteorder == Byteorder.LITTLE:
            byteval = intval.to_bytes(width, "little")
        else:
            byteval = intval.to_bytes(width, "big")

        emulator.write_memory(address, byteval)

    # *** Integer arithmetic constants ***
    #
    # Generalized bitmasks for specific types,
    # determined by their size on this ABI.

    _char_sign_mask: int = 0x80
    _char_inv_mask: int = 0xFF

    _short_sign_mask: int = 0x8000
    _short_inv_mask: int = 0xFFFF

    @property
    @abc.abstractmethod
    def _int_sign_mask(self) -> int:
        # Bitmask covering the sign bit of an int
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _int_inv_mask(self) -> int:
        # Bitmask covering all bits of an int
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _long_sign_mask(self) -> int:
        # Bitmask covering the sign bit of a long
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _long_long_inv_mask(self) -> int:
        # Bitmask covering all bits of a lon
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _long_long_sign_mask(self) -> int:
        # Bitmask covering the sign bit of a long long
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _long_inv_mask(self) -> int:
        # Bitmask covering all bits of a long long
        raise NotImplementedError()

    # Mask for sign-extending 32-bit numbers to 64-bit.
    _int_signext_mask = 0xFFFFFFFF00000000

    # *** Configuration Constants ***
    #
    # It turns out most ABIs follow a generalizable pattern
    # when it comes to passing arguments.
    #
    # The following static fields are the configurations for this pattern.
    # The actual implementation for adding an argument to the signature,
    # and fetching an argument from the emulator
    # are in the functions `add_argument` and `get_argument`.
    #
    # These are separate because they're used by fixed and variadic calls,
    # which are handled separately.

    @property
    @abc.abstractmethod
    def _four_byte_types(self) -> typing.Set[ArgumentType]:
        """Types that are four bytes in this ABI."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _eight_byte_types(self) -> typing.Set[ArgumentType]:
        """Types that are eight bytes in this ABI."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _four_byte_arg_regs(self) -> typing.List[str]:
        """Registers for four-byte arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _eight_byte_arg_regs(self) -> typing.List[str]:
        """Registers for eight-byte arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _soft_float(self) -> bool:
        """Use int regs for fp arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _variadic_soft_float(self) -> bool:
        """Use int regs for fp arguments for variadic args"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _floats_are_doubles(self) -> bool:
        """Floats are actually stored as doubles"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _float_arg_regs(self) -> typing.List[str]:
        """Registers for float arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _double_arg_regs(self) -> typing.List[str]:
        """Registers for double arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _init_stack_offset(self) -> int:
        """Initial offset for stack arguments"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _align_stack(self) -> bool:
        """Align stack for eight-byte values"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _eight_byte_reg_size(self) -> int:
        """Number of registers required for an eight-byte value"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _double_reg_size(self) -> int:
        """Number of registers required for a double value"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _four_byte_stack_size(self) -> int:
        """Size of a four-byte argument on the stack.  Not always four bytes"""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _eight_byte_stack_size(self) -> int:
        """Size of an eight-byte argument on the stack."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _float_stack_size(self) -> int:
        """Size of a float on the stack"""
        raise NotImplementedError()

    @property
    def _double_stack_size(self) -> int:
        """Size of a double on the stack"""
        raise NotImplementedError()


class VariadicContext:
    """Context for extracting variadic arguments

    Variadic functions extend the argument list at runtime.
    Handling this requires dynamically updating the
    list of arguments.  I'd rather do this on a throw-away object
    than on a Model object that will have to reset itself
    for multiple invocations.

    """

    def __init__(self, parent: CStdModel) -> None:
        self.name = parent.name
        self.platform = parent.platform
        self.platdef = parent.platdef

        # Copy the calling convention specification from the parent
        # It's this, or slow the interpreter down by using properties.
        self._int_sign_mask = parent._int_sign_mask
        self._int_inv_mask = parent._int_inv_mask
        self._long_sign_mask = parent._long_sign_mask
        self._long_inv_mask = parent._long_inv_mask
        self._long_long_sign_mask = parent._long_long_sign_mask
        self._long_long_inv_mask = parent._long_long_inv_mask

        self._four_byte_types = parent._four_byte_types
        self._eight_byte_types = parent._eight_byte_types
        self._four_byte_arg_regs = parent._four_byte_arg_regs
        self._eight_byte_arg_regs = parent._eight_byte_arg_regs
        self._soft_float = parent._soft_float or parent._variadic_soft_float
        self._floats_are_doubles = parent._floats_are_doubles
        self._float_arg_regs = parent._float_arg_regs
        self._double_arg_regs = parent._double_arg_regs
        self._init_stack_offset = parent._init_stack_offset
        self._align_stack = parent._align_stack
        self._eight_byte_reg_size = parent._eight_byte_reg_size
        self._double_reg_size = parent._double_reg_size
        self._four_byte_stack_size = parent._four_byte_stack_size
        self._eight_byte_stack_size = parent._eight_byte_stack_size
        self._float_stack_size = parent._float_stack_size
        self._double_stack_size = parent._double_stack_size

        # Copy the parent's current argument state
        self._int_reg_offset = parent._int_reg_offset
        self._fp_reg_offset = parent._fp_reg_offset
        self._stack_offset = parent._stack_offset

        self._on_stack = parent._on_stack.copy()
        self._arg_offset = parent._arg_offset.copy()

    def get_next_argument(
        self, kind: ArgumentType, emulator: emulators.Emulator
    ) -> typing.Union[int, float]:
        """Get the next argument of an assumed type"""
        index = len(self._on_stack)

        add_argument(index, kind, self)
        return get_argument(self, index, kind, emulator)



from ...... import platforms
from .prstatus import PrStatus


class ARM(PrStatus):
    byteorder = platforms.Byteorder.LITTLE

    pr_regs_off = 72
    pr_regs_size = 72

    register_coords = [
        ("r0", 0x0, 4),
        ("r1", 0x4, 4),
        ("r2", 0x8, 4),
        ("r3", 0xC, 4),
        ("r4", 0x10, 4),
        ("r5", 0x14, 4),
        ("r6", 0x18, 4),
        ("r7", 0x1C, 4),
        ("r8", 0x20, 4),
        ("r9", 0x24, 4),
        ("r10", 0x28, 4),
        ("r11", 0x2C, 4),
        ("r12", 0x30, 4),
        ("sp", 0x34, 4),
        ("lr", 0x38, 4),
        ("pc", 0x3C, 4),
        # TODO: This is missing two registers.  One is the CPSR, but I have no idea which.
    ]


class ARMv5T(ARM):
    architecture = platforms.Architecture.ARM_V5T


class ARMv6M(ARM):
    architecture = platforms.Architecture.ARM_V6M


class ARMv6MThumb(ARM):
    architecture = platforms.Architecture.ARM_V6M_THUMB


class ARMv7M(ARM):
    architecture = platforms.Architecture.ARM_V7M


class ARMv7R(ARM):
    architecture = platforms.Architecture.ARM_V7R


class ARMv7A(ARM):
    architecture = platforms.Architecture.ARM_V7A



from .angr import *  # noqa: F401, F403
from .angr import __all__ as __angr__
from .emulator import *  # noqa: F401, F403
from .emulator import __all__ as __emulator__
from .ghidra import *  # noqa: F401, F403
from .ghidra import __all__ as __pcode__
from .unicorn import *  # noqa: F401, F403
from .unicorn import __all__ as __unicorn__

try:
    from .panda import *  # noqa: F401, F403
    from .panda import __all__ as __panda__
except ImportError:
    __panda__ = []

__all__ = __emulator__ + __unicorn__ + __angr__ + __panda__ + __pcode__



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import ArmHFSysVModel


class ArmHFSysVFclose(Fclose, ArmHFSysVModel):
    pass


class ArmHFSysVFeof(Feof, ArmHFSysVModel):
    pass


class ArmHFSysVFerror(Ferror, ArmHFSysVModel):
    pass


class ArmHFSysVFgetc(Fgetc, ArmHFSysVModel):
    pass


class ArmHFSysVFgets(Fgets, ArmHFSysVModel):
    pass


class ArmHFSysVFopen(Fopen, ArmHFSysVModel):
    pass


class ArmHFSysVFprintf(Fprintf, ArmHFSysVModel):
    pass


class ArmHFSysVFputc(Fputc, ArmHFSysVModel):
    pass


class ArmHFSysVFputs(Fputs, ArmHFSysVModel):
    pass


class ArmHFSysVFread(Fread, ArmHFSysVModel):
    pass


class ArmHFSysVFscanf(Fscanf, ArmHFSysVModel):
    pass


class ArmHFSysVFseek(Fseek, ArmHFSysVModel):
    pass


class ArmHFSysVFtell(Ftell, ArmHFSysVModel):
    pass


class ArmHFSysVFwrite(Fwrite, ArmHFSysVModel):
    pass


class ArmHFSysVGetc(Getc, ArmHFSysVModel):
    pass


class ArmHFSysVGetchar(Getchar, ArmHFSysVModel):
    pass


class ArmHFSysVGets(Gets, ArmHFSysVModel):
    pass


class ArmHFSysVPrintf(Printf, ArmHFSysVModel):
    pass


class ArmHFSysVPutc(Putc, ArmHFSysVModel):
    pass


class ArmHFSysVPutchar(Putchar, ArmHFSysVModel):
    pass


class ArmHFSysVPuts(Puts, ArmHFSysVModel):
    pass


class ArmHFSysVRemove(Remove, ArmHFSysVModel):
    pass


class ArmHFSysVRename(Rename, ArmHFSysVModel):
    pass


class ArmHFSysVRewind(Rewind, ArmHFSysVModel):
    pass


class ArmHFSysVScanf(Scanf, ArmHFSysVModel):
    pass


class ArmHFSysVSnprintf(Snprintf, ArmHFSysVModel):
    pass


class ArmHFSysVSprintf(Sprintf, ArmHFSysVModel):
    pass


class ArmHFSysVSscanf(Sscanf, ArmHFSysVModel):
    pass


class ArmHFSysVClearerr(Clearerr, ArmHFSysVModel):
    pass


class ArmHFSysVFflush(Fflush, ArmHFSysVModel):
    pass


class ArmHFSysVFreopen(Freopen, ArmHFSysVModel):
    pass


class ArmHFSysVFgetpos(Fgetpos, ArmHFSysVModel):
    pass


class ArmHFSysVFsetpos(Fsetpos, ArmHFSysVModel):
    pass


class ArmHFSysVTmpfile(Tmpfile, ArmHFSysVModel):
    pass


class ArmHFSysVTmpnam(Tmpnam, ArmHFSysVModel):
    pass


class ArmHFSysVUngetc(Ungetc, ArmHFSysVModel):
    pass


class ArmHFSysVVfprintf(Vfprintf, ArmHFSysVModel):
    pass


class ArmHFSysVVfscanf(Vfscanf, ArmHFSysVModel):
    pass


class ArmHFSysVVprintf(Vprintf, ArmHFSysVModel):
    pass


class ArmHFSysVVscanf(Vscanf, ArmHFSysVModel):
    pass


class ArmHFSysVVsnprintf(Vsnprintf, ArmHFSysVModel):
    pass


class ArmHFSysVVsprintf(Vsprintf, ArmHFSysVModel):
    pass


class ArmHFSysVVsscanf(Vsscanf, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVFclose",
    "ArmHFSysVFeof",
    "ArmHFSysVFerror",
    "ArmHFSysVFgetc",
    "ArmHFSysVFgets",
    "ArmHFSysVFopen",
    "ArmHFSysVFprintf",
    "ArmHFSysVFputc",
    "ArmHFSysVFputs",
    "ArmHFSysVFread",
    "ArmHFSysVFscanf",
    "ArmHFSysVFseek",
    "ArmHFSysVFtell",
    "ArmHFSysVFwrite",
    "ArmHFSysVGetc",
    "ArmHFSysVGetchar",
    "ArmHFSysVGets",
    "ArmHFSysVPrintf",
    "ArmHFSysVPutc",
    "ArmHFSysVPutchar",
    "ArmHFSysVPuts",
    "ArmHFSysVRemove",
    "ArmHFSysVRename",
    "ArmHFSysVRewind",
    "ArmHFSysVScanf",
    "ArmHFSysVSnprintf",
    "ArmHFSysVSprintf",
    "ArmHFSysVSscanf",
    "ArmHFSysVClearerr",
    "ArmHFSysVFflush",
    "ArmHFSysVFreopen",
    "ArmHFSysVFgetpos",
    "ArmHFSysVFsetpos",
    "ArmHFSysVTmpfile",
    "ArmHFSysVTmpnam",
    "ArmHFSysVUngetc",
    "ArmHFSysVVfprintf",
    "ArmHFSysVVfscanf",
    "ArmHFSysVVprintf",
    "ArmHFSysVVscanf",
    "ArmHFSysVVsnprintf",
    "ArmHFSysVVsprintf",
    "ArmHFSysVVsscanf",
]



from ... import platforms, state
from . import cpu


class RISCV64(cpu.CPU):
    """CPU state for riscv64"""

    platform = platforms.Platform(
        platforms.Architecture.RISCV64, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** General-Purpose Registers ***
        # x0 is wired to 0, and aliased as "zero"
        self.x0 = state.FixedRegister("x0", 8, 0)
        self.add(self.x0)
        self.zero = state.RegisterAlias("zero", self.x0, 8, 0)
        self.add(self.zero)
        # x1 acts as the link register
        # NOTE: ra is the official name; lr might be an angr invention.
        self.x1 = state.Register("x1", 8)
        self.add(self.x1)
        self.ra = state.RegisterAlias("ra", self.x1, 8, 0)
        self.add(self.ra)
        # x2 acts as the stack pointer
        self.x2 = state.Register("x2", 8)
        self.add(self.x2)
        self.sp = state.RegisterAlias("sp", self.x2, 8, 0)
        self.add(self.sp)
        # x3 acts as the global pointer
        self.x3 = state.Register("x3", 8)
        self.add(self.x3)
        self.gp = state.RegisterAlias("gp", self.x3, 8, 0)
        self.add(self.gp)
        # x4 acts as the thread pointer
        self.x4 = state.Register("x4", 8)
        self.add(self.x4)
        self.tp = state.RegisterAlias("tp", self.x4, 8, 0)
        self.add(self.tp)
        # x5 is a temporary register
        self.x5 = state.Register("x5", 8)
        self.add(self.x5)
        self.t0 = state.RegisterAlias("t0", self.x5, 8, 0)
        self.add(self.t0)
        # x6 is a temporary register
        self.x6 = state.Register("x6", 8)
        self.add(self.x6)
        self.t1 = state.RegisterAlias("t1", self.x6, 8, 0)
        self.add(self.t1)
        # x7 is a temporary register
        self.x7 = state.Register("x7", 8)
        self.add(self.x7)
        self.t2 = state.RegisterAlias("t2", self.x7, 8, 0)
        self.add(self.t2)
        # x8 is a callee-saved register
        self.x8 = state.Register("x8", 8)
        self.add(self.x8)
        self.s0 = state.RegisterAlias("s0", self.x8, 8, 0)
        self.add(self.s0)
        # x9 is a callee-saved register
        self.x9 = state.Register("x9", 8)
        self.add(self.x9)
        self.s1 = state.RegisterAlias("s1", self.x9, 8, 0)
        self.add(self.s1)
        # x10 is argument 0
        self.x10 = state.Register("x10", 8)
        self.add(self.x10)
        self.a0 = state.RegisterAlias("a0", self.x10, 8, 0)
        self.add(self.a0)
        # x11 is argument 1
        self.x11 = state.Register("x11", 8)
        self.add(self.x11)
        self.a1 = state.RegisterAlias("a1", self.x11, 8, 0)
        self.add(self.a1)
        # x12 is argument 2
        self.x12 = state.Register("x12", 8)
        self.add(self.x12)
        self.a2 = state.RegisterAlias("a2", self.x12, 8, 0)
        self.add(self.a2)
        # x13 is argument 3
        self.x13 = state.Register("x13", 8)
        self.add(self.x13)
        self.a3 = state.RegisterAlias("a3", self.x13, 8, 0)
        self.add(self.a3)
        # x14 is argument 4
        self.x14 = state.Register("x14", 8)
        self.add(self.x14)
        self.a4 = state.RegisterAlias("a4", self.x14, 8, 0)
        self.add(self.a4)
        # x15 is argument 5
        self.x15 = state.Register("x15", 8)
        self.add(self.x15)
        self.a5 = state.RegisterAlias("a5", self.x15, 8, 0)
        self.add(self.a5)
        # x16 is argument 6
        self.x16 = state.Register("x16", 8)
        self.add(self.x16)
        self.a6 = state.RegisterAlias("a6", self.x16, 8, 0)
        self.add(self.a6)
        # x17 is argument 7
        self.x17 = state.Register("x17", 8)
        self.add(self.x17)
        self.a7 = state.RegisterAlias("a7", self.x17, 8, 0)
        self.add(self.a7)
        # x18 is a callee-saved register
        self.x18 = state.Register("x18", 8)
        self.add(self.x18)
        self.s2 = state.RegisterAlias("s2", self.x18, 8, 0)
        self.add(self.s2)
        # x19 is a callee-saved register
        self.x19 = state.Register("x19", 8)
        self.add(self.x19)
        self.s3 = state.RegisterAlias("s3", self.x19, 8, 0)
        self.add(self.s3)
        # x20 is a callee-saved register
        self.x20 = state.Register("x20", 8)
        self.add(self.x20)
        self.s4 = state.RegisterAlias("s4", self.x20, 8, 0)
        self.add(self.s4)
        # x21 is a callee-saved register
        self.x21 = state.Register("x21", 8)
        self.add(self.x21)
        self.s5 = state.RegisterAlias("s5", self.x21, 8, 0)
        self.add(self.s5)
        # x22 is a callee-saved register
        self.x22 = state.Register("x22", 8)
        self.add(self.x22)
        self.s6 = state.RegisterAlias("s6", self.x22, 8, 0)
        self.add(self.s6)
        # x23 is a callee-saved register
        self.x23 = state.Register("x23", 8)
        self.add(self.x23)
        self.s7 = state.RegisterAlias("s7", self.x23, 8, 0)
        self.add(self.s7)
        # x24 is a callee-saved register
        self.x24 = state.Register("x24", 8)
        self.add(self.x24)
        self.s8 = state.RegisterAlias("s8", self.x24, 8, 0)
        self.add(self.s8)
        # x25 is a callee-saved register
        self.x25 = state.Register("x25", 8)
        self.add(self.x25)
        self.s9 = state.RegisterAlias("s9", self.x25, 8, 0)
        self.add(self.s9)
        # x26 is a callee-saved register
        self.x26 = state.Register("x26", 8)
        self.add(self.x26)
        self.s10 = state.RegisterAlias("s10", self.x26, 8, 0)
        self.add(self.s10)
        # x27 is a callee-saved register
        self.x27 = state.Register("x27", 8)
        self.add(self.x27)
        self.s11 = state.RegisterAlias("s11", self.x27, 8, 0)
        self.add(self.s11)
        # x28 is a temporary register
        self.x28 = state.Register("x28", 8)
        self.add(self.x28)
        self.t3 = state.RegisterAlias("t3", self.x28, 8, 0)
        self.add(self.t3)
        # x29 is a temporary register
        self.x29 = state.Register("x29", 8)
        self.add(self.x29)
        self.t4 = state.RegisterAlias("t4", self.x29, 8, 0)
        self.add(self.t4)
        # x30 is a temporary register
        self.x30 = state.Register("x30", 8)
        self.add(self.x30)
        self.t5 = state.RegisterAlias("t5", self.x30, 8, 0)
        self.add(self.t5)
        # x31 is a temporary register
        self.x31 = state.Register("x31", 8)
        self.add(self.x31)
        self.t6 = state.RegisterAlias("t6", self.x31, 8, 0)
        self.add(self.t6)

        # *** Program Counter ***
        self.pc = state.Register("pc", 8)
        self.add(self.pc)

        # *** Floating-Point Registers ***
        # f0 is a temporary register
        self.f0 = state.Register("f0", 8)
        self.add(self.f0)
        self.ft0 = state.RegisterAlias("ft0", self.f0, 8, 0)
        self.add(self.ft0)
        # f1 is a temporary register
        self.f1 = state.Register("f1", 8)
        self.add(self.f1)
        self.ft1 = state.RegisterAlias("ft1", self.f1, 8, 0)
        self.add(self.ft1)
        # f2 is a temporary register
        self.f2 = state.Register("f2", 8)
        self.add(self.f2)
        self.ft2 = state.RegisterAlias("ft2", self.f2, 8, 0)
        self.add(self.ft2)
        # f3 is a temporary register
        self.f3 = state.Register("f3", 8)
        self.add(self.f3)
        self.ft3 = state.RegisterAlias("ft3", self.f3, 8, 0)
        self.add(self.ft3)
        # f4 is a temporary register
        self.f4 = state.Register("f4", 8)
        self.add(self.f4)
        self.ft4 = state.RegisterAlias("ft4", self.f4, 8, 0)
        self.add(self.ft4)
        # f5 is a temporary register
        self.f5 = state.Register("f5", 8)
        self.add(self.f5)
        self.ft5 = state.RegisterAlias("ft5", self.f5, 8, 0)
        self.add(self.ft5)
        # f6 is a temporary register
        self.f6 = state.Register("f6", 8)
        self.add(self.f6)
        self.ft6 = state.RegisterAlias("ft6", self.f6, 8, 0)
        self.add(self.ft6)
        # f7 is a temporary register
        self.f7 = state.Register("f7", 8)
        self.add(self.f7)
        self.ft7 = state.RegisterAlias("ft7", self.f7, 8, 0)
        self.add(self.ft7)
        # f8 is a callee saved register
        self.f8 = state.Register("f8", 8)
        self.add(self.f8)
        self.fs0 = state.RegisterAlias("fs0", self.f8, 8, 0)
        self.add(self.fs0)
        # f9 is a callee saved register
        self.f9 = state.Register("f9", 8)
        self.add(self.f9)
        self.fs1 = state.RegisterAlias("fs1", self.f9, 8, 0)
        self.add(self.fs1)
        # f10 is argument 0
        self.f10 = state.Register("f10", 8)
        self.add(self.f10)
        self.fa0 = state.RegisterAlias("fa0", self.f10, 8, 0)
        self.add(self.fa0)
        # f11 is argument 1
        self.f11 = state.Register("f11", 8)
        self.add(self.f11)
        self.fa1 = state.RegisterAlias("fa1", self.f11, 8, 0)
        self.add(self.fa1)
        # f12 is argument 2
        self.f12 = state.Register("f12", 8)
        self.add(self.f12)
        self.fa2 = state.RegisterAlias("fa2", self.f12, 8, 0)
        self.add(self.fa2)
        # f13 is argument 3
        self.f13 = state.Register("f13", 8)
        self.add(self.f13)
        self.fa3 = state.RegisterAlias("fa3", self.f13, 8, 0)
        self.add(self.fa3)
        # f14 is argument 4
        self.f14 = state.Register("f14", 8)
        self.add(self.f14)
        self.fa4 = state.RegisterAlias("fa4", self.f14, 8, 0)
        self.add(self.fa4)
        # f15 is argument 5
        self.f15 = state.Register("f15", 8)
        self.add(self.f15)
        self.fa5 = state.RegisterAlias("fa5", self.f15, 8, 0)
        self.add(self.fa5)
        # f16 is argument 6
        self.f16 = state.Register("f16", 8)
        self.add(self.f16)
        self.fa6 = state.RegisterAlias("fa6", self.f16, 8, 0)
        self.add(self.fa6)
        # f7 is argument 7
        self.f17 = state.Register("f17", 8)
        self.add(self.f17)
        self.fa7 = state.RegisterAlias("fa7", self.f17, 8, 0)
        self.add(self.fa7)
        # f18 is a callee-saved register
        self.f18 = state.Register("f18", 8)
        self.add(self.f18)
        self.fs2 = state.RegisterAlias("fs2", self.f18, 8, 0)
        self.add(self.fs2)
        # f19 is a callee-saved register
        self.f19 = state.Register("f19", 8)
        self.add(self.f19)
        self.fs3 = state.RegisterAlias("fs3", self.f19, 8, 0)
        self.add(self.fs3)
        # f20 is a callee-saved register
        self.f20 = state.Register("f20", 8)
        self.add(self.f20)
        self.fs4 = state.RegisterAlias("fs4", self.f20, 8, 0)
        self.add(self.fs4)
        # f21 is a callee-saved register
        self.f21 = state.Register("f21", 8)
        self.add(self.f21)
        self.fs5 = state.RegisterAlias("fs5", self.f21, 8, 0)
        self.add(self.fs5)
        # f22 is a callee-saved register
        self.f22 = state.Register("f22", 8)
        self.add(self.f22)
        self.fs6 = state.RegisterAlias("fs6", self.f22, 8, 0)
        self.add(self.fs6)
        # f23 is a callee-saved register
        self.f23 = state.Register("f23", 8)
        self.add(self.f23)
        self.fs7 = state.RegisterAlias("fs7", self.f23, 8, 0)
        self.add(self.fs7)
        # f24 is a callee-saved register
        self.f24 = state.Register("f24", 8)
        self.add(self.f24)
        self.fs8 = state.RegisterAlias("fs8", self.f24, 8, 0)
        self.add(self.fs8)
        # f25 is a callee-saved register
        self.f25 = state.Register("f25", 8)
        self.add(self.f25)
        self.fs9 = state.RegisterAlias("fs9", self.f25, 8, 0)
        self.add(self.fs9)
        # f26 is a callee-saved register
        self.f26 = state.Register("f26", 8)
        self.add(self.f26)
        self.fs10 = state.RegisterAlias("fs10", self.f26, 8, 0)
        self.add(self.fs10)
        # f27 is a callee-saved register
        self.f27 = state.Register("f27", 8)
        self.add(self.f27)
        self.fs11 = state.RegisterAlias("fs11", self.f27, 8, 0)
        self.add(self.fs11)
        # f28 is a temporary register
        self.f28 = state.Register("f28", 8)
        self.add(self.f28)
        self.ft8 = state.RegisterAlias("ft8", self.f28, 8, 0)
        self.add(self.ft8)
        # f29 is a temporary register
        self.f29 = state.Register("f29", 8)
        self.add(self.f29)
        self.ft9 = state.RegisterAlias("ft9", self.f29, 8, 0)
        self.add(self.ft9)
        # f30 is a temporary register
        self.f30 = state.Register("f30", 8)
        self.add(self.f30)
        self.ft10 = state.RegisterAlias("ft10", self.f30, 8, 0)
        self.add(self.ft10)
        # f31 is a temporary register
        self.f31 = state.Register("f31", 8)
        self.add(self.f31)
        self.ft11 = state.RegisterAlias("ft11", self.f31, 8, 0)
        self.add(self.ft11)

        # *** Vector Registers ***
        # NOTE: These exist, but are not supported

        # *** Control and Status Registers ***
        # NOTE: These exist, but aren't supported.



import typing

from .... import platforms
from . import stack


class MIPS64Stack(stack.DescendingStack):
    """A stack for a MIPS 64-bit CPU"""

    def get_alignment(self) -> int:
        return 4

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for MIPS64")


class MIPS64BEStack(MIPS64Stack):
    """A stack for a big-endian MIPS 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.MIPS64, platforms.Byteorder.BIG
    )


class MIPS64ELStack(MIPS64Stack):
    """A stack for a little-endian MIPS 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.MIPS64, platforms.Byteorder.LITTLE
    )



from ....c99.signal import Raise, Signal
from ..systemv import MIPS64SysVModel


class MIPS64SysVRaise(Raise, MIPS64SysVModel):
    pass


class MIPS64SysVSignal(Signal, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVRaise",
    "MIPS64SysVSignal",
]



import typing

from .... import platforms
from . import stack


class PowerPCStack(stack.DescendingStack):
    """A stack for a PPC 32-bit CPU"""

    def get_alignment(self) -> int:
        return 4

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for PowerPC")


class PowerPC32Stack(PowerPCStack):
    """A stack for a PPC 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.POWERPC32, platforms.Byteorder.BIG
    )


class PowerPC64Stack(PowerPCStack):
    """A stack for a PPC 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.POWERPC64, platforms.Byteorder.BIG
    )



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import ArmELSysVModel


class ArmELSysVTime(Time, ArmELSysVModel):
    pass


class ArmELSysVLocaltime(Localtime, ArmELSysVModel):
    pass


class ArmELSysVGmtime(Gmtime, ArmELSysVModel):
    pass


class ArmELSysVCtime(Ctime, ArmELSysVModel):
    pass


class ArmELSysVAsctime(Asctime, ArmELSysVModel):
    pass


class ArmELSysVStrftime(Strftime, ArmELSysVModel):
    pass


class ArmELSysVDifftime(Difftime, ArmELSysVModel):
    pass


class ArmELSysVMktime(Mktime, ArmELSysVModel):
    pass


class ArmELSysVClock(Clock, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVTime",
    "ArmELSysVLocaltime",
    "ArmELSysVGmtime",
    "ArmELSysVCtime",
    "ArmELSysVAsctime",
    "ArmELSysVStrftime",
    "ArmELSysVDifftime",
    "ArmELSysVMktime",
    "ArmELSysVClock",
]



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import AArch64SysVModel


class AArch64SysVTime(Time, AArch64SysVModel):
    pass


class AArch64SysVLocaltime(Localtime, AArch64SysVModel):
    pass


class AArch64SysVGmtime(Gmtime, AArch64SysVModel):
    pass


class AArch64SysVCtime(Ctime, AArch64SysVModel):
    pass


class AArch64SysVAsctime(Asctime, AArch64SysVModel):
    pass


class AArch64SysVStrftime(Strftime, AArch64SysVModel):
    pass


class AArch64SysVDifftime(Difftime, AArch64SysVModel):
    pass


class AArch64SysVMktime(Mktime, AArch64SysVModel):
    pass


class AArch64SysVClock(Clock, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVTime",
    "AArch64SysVLocaltime",
    "AArch64SysVGmtime",
    "AArch64SysVCtime",
    "AArch64SysVAsctime",
    "AArch64SysVStrftime",
    "AArch64SysVDifftime",
    "AArch64SysVMktime",
    "AArch64SysVClock",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import MIPSELSysVModel


class MIPSELSysVMemcpy(Memcpy, MIPSELSysVModel):
    pass


class MIPSELSysVMemmove(Memmove, MIPSELSysVModel):
    pass


class MIPSELSysVStrcpy(Strcpy, MIPSELSysVModel):
    pass


class MIPSELSysVStrncpy(Strncpy, MIPSELSysVModel):
    pass


class MIPSELSysVStrcat(Strcat, MIPSELSysVModel):
    pass


class MIPSELSysVStrncat(Strncat, MIPSELSysVModel):
    pass


class MIPSELSysVMemcmp(Memcmp, MIPSELSysVModel):
    pass


class MIPSELSysVStrncmp(Strncmp, MIPSELSysVModel):
    pass


class MIPSELSysVStrcmp(Strcmp, MIPSELSysVModel):
    pass


class MIPSELSysVStrcoll(Strcoll, MIPSELSysVModel):
    pass


class MIPSELSysVStrxfrm(Strxfrm, MIPSELSysVModel):
    pass


class MIPSELSysVMemchr(Memchr, MIPSELSysVModel):
    pass


class MIPSELSysVStrchr(Strchr, MIPSELSysVModel):
    pass


class MIPSELSysVStrcspn(Strcspn, MIPSELSysVModel):
    pass


class MIPSELSysVStrpbrk(Strpbrk, MIPSELSysVModel):
    pass


class MIPSELSysVStrrchr(Strrchr, MIPSELSysVModel):
    pass


class MIPSELSysVStrspn(Strspn, MIPSELSysVModel):
    pass


class MIPSELSysVStrstr(Strstr, MIPSELSysVModel):
    pass


class MIPSELSysVStrtok(Strtok, MIPSELSysVModel):
    pass


class MIPSELSysVMemset(Memset, MIPSELSysVModel):
    pass


class MIPSELSysVStrerror(Strerror, MIPSELSysVModel):
    pass


class MIPSELSysVStrlen(Strlen, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVMemcpy",
    "MIPSELSysVMemmove",
    "MIPSELSysVStrcpy",
    "MIPSELSysVStrncpy",
    "MIPSELSysVStrcat",
    "MIPSELSysVStrncat",
    "MIPSELSysVMemcmp",
    "MIPSELSysVStrncmp",
    "MIPSELSysVStrcmp",
    "MIPSELSysVStrcoll",
    "MIPSELSysVStrxfrm",
    "MIPSELSysVMemchr",
    "MIPSELSysVStrchr",
    "MIPSELSysVStrcspn",
    "MIPSELSysVStrpbrk",
    "MIPSELSysVStrrchr",
    "MIPSELSysVStrspn",
    "MIPSELSysVStrstr",
    "MIPSELSysVStrtok",
    "MIPSELSysVMemset",
    "MIPSELSysVStrerror",
    "MIPSELSysVStrlen",
]



from ....c99.signal import Raise, Signal
from ..systemv import PowerPCSysVModel


class PowerPCSysVRaise(Raise, PowerPCSysVModel):
    pass


class PowerPCSysVSignal(Signal, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVRaise",
    "PowerPCSysVSignal",
]



from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef

# NOTE: Xtensa is designed to be an extensible language
# There are options for Xtensa chips to implement almost any
# ISA feature you can bally well think of,
# some of which are closed-source and closed-spec.
#
# Support for any specific extension depends on the emulator.


class Xtensa(PlatformDef):
    architecture = Architecture.XTENSA
    byteorder = Byteorder.LITTLE

    address_size = 4

    conditional_branch_mnemonics = {
        # Compare to zero and branch
        "beqz",
        "bnez",
        "bgez",
        "bltz",
        # Compare to immediate and branch
        "beqi",
        "bnei",
        "bgei",
        "blti",
        # Compare to unsigned immediate and branch
        "bgeui",
        "bltui",
        # Compare to register and branch
        "beq",
        "bne",
        "bge",
        "blt",
        # Compare to unsigned register and branch
        "bgeu",
        "bltu",
        # Bit test versus immediate and branch
        "bbci",
        "bbsi",
        # Bit test versus register and branch
        "bbc",
        "bbs",
        # Test against bitmask and branch
        "bany",
        "bnone",
        "ball",
        "bnall",
    }

    # Xtensa core has no comparison operations;
    # they're built into the branch instructions.
    compare_mnemonics = set()

    # NOTE: Capstone does not yet support xtensa
    # It looks like it's either planned, or present in a newer version than we use.
    capstone_arch = -1
    capstone_mode = -1

    pc_register = "pc"
    sp_register = "sp"

    # Special registers:
    # - a0 is the default link register
    # - a1 is the stack pointer
    general_purpose_registers = [f"a{i}" for i in range(2, 16)]

    registers = {
        # *** General Purpose Registers ***
        # a0 is also the default link register, but it doesn't get an alias
        "a0": RegisterDef(name="a0", size=4),
        # a1 is also the stack pointer
        "a1": RegisterDef(name="a1", size=4),
        "sp": RegisterAliasDef(name="sp", parent="a1", size=4, offset=0),
        "a2": RegisterDef(name="a2", size=4),
        "a3": RegisterDef(name="a3", size=4),
        "a4": RegisterDef(name="a4", size=4),
        "a5": RegisterDef(name="a5", size=4),
        "a6": RegisterDef(name="a6", size=4),
        "a7": RegisterDef(name="a7", size=4),
        "a8": RegisterDef(name="a8", size=4),
        "a9": RegisterDef(name="a9", size=4),
        "a10": RegisterDef(name="a10", size=4),
        "a11": RegisterDef(name="a11", size=4),
        "a12": RegisterDef(name="a12", size=4),
        "a13": RegisterDef(name="a13", size=4),
        "a14": RegisterDef(name="a14", size=4),
        "a15": RegisterDef(name="a15", size=4),
        # *** Program Counter ***
        "pc": RegisterDef(name="pc", size=4),
        # *** Shift Amount Register ***
        # This thing is actually 6 bits.
        "sar": RegisterDef(name="sar", size=4),
    }



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import PowerPCSysVModel


class PowerPCSysVBsdSignal(BsdSignal, PowerPCSysVModel):
    pass


class PowerPCSysVKill(Kill, PowerPCSysVModel):
    pass


class PowerPCSysVKillpg(Killpg, PowerPCSysVModel):
    pass


class PowerPCSysVPthreadKill(PthreadKill, PowerPCSysVModel):
    pass


class PowerPCSysVPthreadSigmask(PthreadSigmask, PowerPCSysVModel):
    pass


class PowerPCSysVSigaction(Sigaction, PowerPCSysVModel):
    pass


class PowerPCSysVSigaddset(Sigaddset, PowerPCSysVModel):
    pass


class PowerPCSysVSigaltstack(Sigaltstack, PowerPCSysVModel):
    pass


class PowerPCSysVSigdelset(Sigdelset, PowerPCSysVModel):
    pass


class PowerPCSysVSigemptyset(Sigemptyset, PowerPCSysVModel):
    pass


class PowerPCSysVSigfillset(Sigfillset, PowerPCSysVModel):
    pass


class PowerPCSysVSighold(Sighold, PowerPCSysVModel):
    pass


class PowerPCSysVSigignore(Sigignore, PowerPCSysVModel):
    pass


class PowerPCSysVSiginterrupt(Siginterrupt, PowerPCSysVModel):
    pass


class PowerPCSysVSigismember(Sigismember, PowerPCSysVModel):
    pass


class PowerPCSysVSigpause(Sigpause, PowerPCSysVModel):
    pass


class PowerPCSysVSigpending(Sigpending, PowerPCSysVModel):
    pass


class PowerPCSysVSigprocmask(Sigprocmask, PowerPCSysVModel):
    pass


class PowerPCSysVSigqueue(Sigqueue, PowerPCSysVModel):
    pass


class PowerPCSysVSigrelse(Sigrelse, PowerPCSysVModel):
    pass


class PowerPCSysVSigset(Sigset, PowerPCSysVModel):
    pass


class PowerPCSysVSigsuspend(Sigsuspend, PowerPCSysVModel):
    pass


class PowerPCSysVSigtimedwait(Sigtimedwait, PowerPCSysVModel):
    pass


class PowerPCSysVSigwait(Sigwait, PowerPCSysVModel):
    pass


class PowerPCSysVSigwaitinfo(Sigwaitinfo, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVBsdSignal",
    "PowerPCSysVKill",
    "PowerPCSysVKillpg",
    "PowerPCSysVPthreadKill",
    "PowerPCSysVPthreadSigmask",
    "PowerPCSysVSigaction",
    "PowerPCSysVSigaddset",
    "PowerPCSysVSigaltstack",
    "PowerPCSysVSigdelset",
    "PowerPCSysVSigemptyset",
    "PowerPCSysVSigfillset",
    "PowerPCSysVSighold",
    "PowerPCSysVSigignore",
    "PowerPCSysVSiginterrupt",
    "PowerPCSysVSigismember",
    "PowerPCSysVSigpause",
    "PowerPCSysVSigpending",
    "PowerPCSysVSigprocmask",
    "PowerPCSysVSigqueue",
    "PowerPCSysVSigrelse",
    "PowerPCSysVSigset",
    "PowerPCSysVSigsuspend",
    "PowerPCSysVSigtimedwait",
    "PowerPCSysVSigwait",
    "PowerPCSysVSigwaitinfo",
]



import logging
import typing

import lief

from ....exceptions import ConfigurationError
from ....platforms import Architecture, Byteorder, Platform
from ....utils import RangeCollection
from ...state import BytesValue
from ..code import Executable
from .rela import ElfRelocator
from .structs import ElfRela, ElfSymbol

log = logging.getLogger(__name__)

# ELF machine values
# See /usr/include/elf.h for the complete list
EM_386 = 3  # Intel 80386
EM_MIPS = 8  # MIPS; all kinds
EM_PPC = 20  # PowerPC 32-bit
EM_PPC64 = 21  # PowerPC 64-bit
EM_ARM = 40  # ARM 32-bit
EM_X86_64 = 62  # AMD/Intel x86-64
EM_XTENSA = 94  # Xtensa
EM_AARCH64 = 183  # ARM v9, or AARCH64
EM_RISCV = 243  # RISC-V
EM_LOONGARCH = 258  # LoongArch

# ARM-specific flag values
EF_ARM_VFP_FLOAT = 0x400
EF_ARM_SOFT_FLOAT = 0x200
EF_ARM_EABI_VER5 = 0x05000000

# Program header types
PT_NULL = 0  # Empty/unused program header
PT_LOAD = 1  # Describes loadable program segment
PT_DYNAMIC = 2  # Points to dynamic linking metadata
PT_INTERP = 3  # Points to program interpreter
PT_NOTE = 4  # Points to auxiliary information
PT_SHLIB = 5  # Reserved value; I think it's unused
PT_PHDR = 6  # Points to program header table
PT_TLS = 7  # Indicates need for thread-local storage
PT_LOOS = 0x60000000  # Start of OS-specific types
PT_GNU_EH_FRAME = 0x6474E550  # GNU-specific: Points to exception handler segment
PT_GNU_STACK = 0x6474E551  # GNU-specific: Describes stack permissions
PT_GNU_RELRO = 0x6474E552  # GNU-specific: Describes read-only after relocation segment
PT_GNU_PROPERTY = 0x6474E553  # GNU-specific: Points to GNU property
PT_HIOS = 0x6FFFFFFF  # End of OS-specific types
PT_LOPROC = 0x70000000  # Start of processor-specific types
PT_HIPROC = 0x7FFFFFFF  # End of processor-specific types

# Program header flags
PF_X = 0x1  # Segment is executable
PF_W = 0x2  # Segment is writable
PF_R = 0x4  # Segment is readable

# Universal dynamic tag values
DT_PLTGOT = 0x3

# MIPS-specific dynamic tag values
DT_MIPS_LOCAL_GOTNO = 0x7000000A
DT_MIPS_GOTSYM = 0x70000013

# MIPS-specific relocation type
R_MIPS_32 = 2
R_MIPS_64 = 18


class ElfExecutable(Executable):
    """Executable loaded from an ELF

    This loads a single ELF file into a SmallWorld memory object.
    It performs no relocation or any other initialization,
    just maps the file into memory as the kernel intended.

    Arguments:
        file: File-like object containing the image
        platform: Optional platform; used for header verification
        ignore_platform: Do not try to ID or verify platform from headers
        user_base: Optional user-specified base address
        page_size: System page size
    """

    def __init__(
        self,
        file: typing.BinaryIO,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
        user_base: typing.Optional[int] = None,
        page_size: int = 0x1000,
    ):
        # Initialize with null address and size;
        # we will update these later.
        super().__init__(0, 0)
        self.platform = platform
        self.bounds: RangeCollection = RangeCollection()
        self._page_size = page_size
        self._user_base = user_base
        self._file_base = 0

        # Lief struct.
        # We can't keep it around forever, since it's not copyable.
        self._elf: typing.Optional[typing.Any] = None

        # Initialize dynamic tag info
        self._dtags: typing.Dict[int, int] = dict()

        # Initialize symbol info
        self._dynamic_symbols: typing.List[ElfSymbol] = list()
        self._static_symbols: typing.List[ElfSymbol] = list()
        self._dynamic_relas: typing.List[ElfRela] = list()
        self._static_relas: typing.List[ElfRela] = list()
        self._syms_by_name: typing.Dict[str, typing.List[ElfSymbol]] = dict()
        self._relocator: typing.Optional[ElfRelocator] = None

        # Read the entire image out of the file.
        image = file.read()

        # Use lief to check if this is an ELF.
        # NOTE: For some reason, this takes list(int), not bytes
        if not lief.is_elf(list(image)):
            raise ConfigurationError("Image is not an ELF")

        # Use lief to parse the ELF.
        # NOTE: For some reason, this takes list(int), not bytes
        # NOTE: lief objects aren't deep-copyable.
        # I'd love to keep `elf` around for later use, but I can't.
        elf = lief.ELF.parse(list(image))
        if elf is None:
            raise ConfigurationError("Failed parsing ELF")

        # Save the lief struct
        self._elf = elf

        # Extract the file header
        ehdr = elf.header
        if ehdr is None:
            raise ConfigurationError("Failed extracting ELF header")

        # Check machine compatibility
        if not ignore_platform:
            hdr_platform = self._platform_for_ehdr(elf)
            if self.platform is not None:
                if self.platform != hdr_platform:
                    raise ConfigurationError(
                        "Platform mismatch: "
                        f"specified {self.platform}, but got {hdr_platform} from header"
                    )
            else:
                self.platform = hdr_platform

        if self.platform is not None:
            # If we have a platform, we can relocate
            self._relocator = ElfRelocator.for_platform(self.platform)

        # Figure out if this file is loadable.
        # If there are program headers, it's loadable.
        # This doesn't support .ko files, are relocatable objects,
        # and have no program headers.
        if ehdr.program_header_offset == 0:
            # No program headers; not a loadable ELF.
            raise ConfigurationError("ELF is not loadable")

        if ehdr.program_header_offset >= len(image):
            # Obviously-invalid program headers
            raise ConfigurationError(
                f"Invalid program header offset {hex(ehdr.program_header_offset)}"
            )

        # Determine the file base address.
        self._file_base = elf.imagebase
        self._determine_base()

        for phdr in elf.segments:
            log.debug(f"{phdr}")
            phdr_type = phdr.type.value & 0xFFFFFFFF
            if phdr_type == PT_LOAD:
                # Loadable segment
                # Map its data into memory
                self._map_segment(phdr, image)
            elif phdr_type == PT_DYNAMIC:
                # Dynamic linking metadata.
                # This ELF needs dynamic linking
                log.info("Program includes dynamic linking metadata")
            elif phdr_type == PT_INTERP:
                # Program interpreter
                # This completely changes how program loading works.
                # Whether you care is a different matter.
                interp = image[phdr.file_offset : phdr.file_offset + phdr.physical_size]
                log.info(f"Program specifies interpreter {interp!r}")
            elif phdr_type == PT_NOTE:
                # Auxiliary information
                # Possibly useful for comparing machine/OS type.
                pass
            elif phdr_type == PT_PHDR:
                # Program header self-reference
                # Useful for the dynamic linker, but not for us
                pass
            elif phdr_type == PT_TLS:
                # TLS Segment
                # Your analysis is about to get nasty :(
                log.info("Program includes thread-local storage")
            elif phdr_type == PT_GNU_EH_FRAME:
                # Exception handler frame.
                # GCC puts one of these in everything.  Do we care?
                pass
            elif phdr_type == PT_GNU_STACK:
                # Stack executability
                # If this is missing, assume executable stack
                log.info("Program specifies stack permissions")
            elif phdr_type == PT_GNU_RELRO:
                # Read-only after relocation
                # Only the dynamic linker should write this data.
                log.info("Program specifies RELRO data")
            elif phdr_type == PT_GNU_PROPERTY:
                # GNU property segment
                # Contains extra metadata which I'm not sure anything uses
                pass
            elif phdr_type >= PT_LOOS and phdr_type <= PT_HIOS:
                # Unknown OS-specific program header
                # Either this is a weird ISA that extends the generic GNU ABI,
                # or this isn't a Linux ELF.
                log.warn(f"Unknown OS-specific program header: {phdr_type:08x}")
            elif phdr_type >= PT_LOPROC and phdr_type <= PT_HIPROC:
                # Unknown machine-specific program header
                # This is probably a non-Intel ISA.
                # Most of these are harmless, serving to tell the RTLD
                # where to find machine-specific metadata
                log.warn(f"Unknown machine-specific program header: {phdr_type:08x}")
            else:
                # Unknown program header outside the allowed custom ranges
                log.warn(f"Invalid program header: {phdr_type:08x}")

        # Compute the final total capacity
        for offset, value in self.items():
            self.size = max(self.size, offset + value.get_size())

        # Determine if the file specifies an entrypoint
        if elf.entrypoint is not None and elf.entrypoint != 0:
            # Check if the entrypoint is valid (falls within the image)
            entrypoint = self._rebase_file(elf.entrypoint)
            if not self.bounds.contains_value(entrypoint):
                if (
                    self.platform is not None
                    and self.platform.architecture == Architecture.POWERPC64
                ):
                    # NOTE: PowerPC64's ABI is trippy.
                    # It uses "function descriptor" structs instead of
                    # simple function pointers.
                    # Thus, the entrypoint points to something in the data section.
                    log.warn("Entrypoint for PowerPC64 file is not a code address")
                else:
                    raise ConfigurationError(
                        f"Invalid entrypoint address {hex(entrypoint)}"
                    )
            else:
                self.entrypoint = entrypoint
        else:
            # No entrypoint specified,
            # Or this is PowerPC64 and the entrypoint is gibberish
            self.entrypoint = None

        # Organize dynamic tags
        self._extract_dtags(elf)

        # Organize symbols for later relocation
        self._extract_symbols(elf)

    def _platform_for_ehdr(self, elf):
        # Determine byteorder.  This bit's easy
        if elf.header.identity_data.value == 1:
            # LSB byteorder
            byteorder = Byteorder.LITTLE
        elif elf.header.identity_data.value == 2:
            # MSB byteorder
            byteorder = Byteorder.BIG
        else:
            raise ConfigurationError(
                f"Unknown value of ei_data: {hex(elf.header.identity_data.value)}"
            )

        # Determine arch/mode.  This bit's harder.
        if elf.header.machine_type.value == EM_X86_64:
            # amd64
            architecture = Architecture.X86_64
        elif elf.header.machine_type.value == EM_AARCH64:
            # aarch64
            architecture = Architecture.AARCH64
        elif elf.header.machine_type.value == EM_386:
            # i386
            architecture = Architecture.X86_32
        elif elf.header.machine_type.value == EM_ARM:
            # Some kind of arm32
            flags = set(map(lambda x: x.value & 0xFFFFFFFF, elf.header.flags_list))

            if EF_ARM_EABI_VER5 in flags and EF_ARM_SOFT_FLOAT in flags:
                # This is either ARMv5T or some kind of ARMv6.
                # We're currently assuming v5T, but this isn't always correct.
                architecture = Architecture.ARM_V5T
            elif EF_ARM_EABI_VER5 in flags and EF_ARM_VFP_FLOAT in flags:
                # This is ARMv7a, as built by gcc.
                architecture = Architecture.ARM_V7A
            else:
                # The file might be a core dump or some uncommon flavor of ARM.
                # Fallback to a default if flags are missing.
                log.warning(
                    f"No recognized ARM flags found. flags={list(map(hex, flags))}, "
                    "defaulting to ARM_V7A."
                )
                architecture = Architecture.ARM_V7A
        elif elf.header.machine_type.value == EM_LOONGARCH:
            if elf.header.identity_class.value == 1:
                # 32-bit elf
                raise ConfigurationError("LoongArch32 not supported")
            elif elf.header.identity_class.value == 2:
                # 64-bit elf
                architecture = Architecture.LOONGARCH64
            else:
                raise ConfigurationError(
                    f"Unknown value of ei_class: {hex(elf.header.identity_class.value)}"
                )
        elif elf.header.machine_type.value == EM_MIPS:
            # Some kind of mips.
            # TODO: There are more parameters than just word size
            if elf.header.identity_class.value == 1:
                # 32-bit ELF
                architecture = Architecture.MIPS32
            elif elf.header.identity_class.value == 2:
                architecture = Architecture.MIPS64
            else:
                raise ConfigurationError(
                    f"Unknown value of ei_class: {hex(elf.header.identity_class.value)}"
                )
        elif elf.header.machine_type.value == EM_PPC:
            # PowerPC 32-bit
            architecture = Architecture.POWERPC32
        elif elf.header.machine_type.value == EM_PPC64:
            # PowerPC 64-bit
            architecture = Architecture.POWERPC64
        elif elf.header.machine_type.value == EM_RISCV:
            # RISC-V
            if elf.header.identity_class.value == 1:
                raise ConfigurationError("RISC-V 32-bit isn't supported")
            elif elf.header.identity_class.value == 2:
                architecture = Architecture.RISCV64
            else:
                raise ConfigurationError(
                    f"Unknown value of ei_class: {hex(elf.header.identity_class.value)}"
                )
        elif elf.header.machine_type.value == EM_XTENSA:
            architecture = Architecture.XTENSA
        else:
            raise ConfigurationError(
                f"Unknown value of e_machine: {hex(elf.header.machine_type.value)}"
            )
        return Platform(architecture, byteorder)

    def _determine_base(self):
        # Determine the base address of this image
        #
        # Normally, the loader respects the wishes of the file;
        # an external base address is only used if the image doesn't define one.
        #
        # Here, the user doesn't know the image layout ahead of time,
        # and may need to adjust their environment to fit,
        # so a user-provided base is given equal weight.
        #
        # If both base addresses are specified,
        # or neither is specified, it's a configuratione rror.

        if self._user_base is None:
            # No user base requested
            if self._file_base == 0:
                # No file base defined.
                # Need the user to provide one
                raise ConfigurationError(
                    "No base address provided for position-independent ELF image"
                )
            else:
                self.address = self._file_base
        else:
            # User base requested
            if self._file_base == 0:
                # No file base requested; we are okay with this.
                self.address = self._user_base
            elif self._user_base == self._file_base:
                # Everyone requested the same base address, so we're okay.
                self.address = self._user_base
            else:
                # File base is defined.
                # We (probably) cannot move the image without problems.
                raise ConfigurationError("Base address defined for fixed-position ELF")
        log.info(f"Address: {self.address:x}")

    def _rebase_file(self, val: int):
        # Rebase an offset from file-relative to image-relative
        return val - self._file_base + self.address

    def _page_align(self, val: int, up: bool = True):
        # Align an address to a page boundary
        # There are a number of cases where ELF files are imprecise;
        # they rely on the kernel/libc to map things at page-aligned addresses.
        if up:
            val += self._page_size - 1
        return (val // self._page_size) * self._page_size

    def _map_segment(self, phdr, image):
        # Compute segment boundaries
        seg_start = self._page_align(phdr.file_offset, up=False)
        seg_end = self._page_align(phdr.file_offset + phdr.physical_size)
        seg_addr = self._page_align(self._rebase_file(phdr.virtual_address), up=False)
        seg_size = self._page_align(phdr.virtual_size + (phdr.file_offset - seg_start))

        log.debug("Mapping: ")
        log.debug(
            f"    f: [ {seg_start:012x} -> {seg_end:012x} ] ( {seg_end - seg_start:x} bytes )"
        )
        log.debug(
            f"    m: [ {seg_addr:012x} -> {seg_addr + seg_size:012x} ] ( {seg_size:x} bytes )"
        )

        # Extract segment data
        seg_data = image[seg_start:seg_end]
        if len(seg_data) < seg_size:
            # Segment is shorter than is available from the file;
            # this will get zero-padded.
            pad_size = seg_size - len(seg_data)
            log.debug(f"Padding segment by {pad_size} bytes")
            seg_data += b"\0" * pad_size
        if len(seg_data) != seg_size:
            raise ConfigurationError(
                f"Expected segment of size {seg_size}, but got {len(seg_data)}"
            )
        if (phdr.flags.value & PF_X) != 0:
            # This is a code segment; add it to program bounds
            self.bounds.add_range((seg_addr, seg_addr + seg_size))

        # Add the segment to the memory map
        seg_value = BytesValue(seg_data, None)
        assert (
            seg_value._size == seg_size
        ), f"Expected {seg_size:x} bytes, got {seg_value._size:x}"
        self[seg_addr - self.address] = seg_value

    def _extract_dtags(self, elf):
        for dt in elf.dynamic_entries:
            self._dtags[dt.tag.value & 0xFFFFFFFF] = dt.value

    def _extract_symbols(self, elf):
        lief_to_elf = dict()

        # Figure out the base address
        # TODO: Currently, this only handles PIC or Non-PIC
        # It will be wrong for .o and .ko
        if self._file_base is not None and self._file_base != 0:
            # This is a non-PIC binary.
            # Dollars to ducats the symbol is absolute.
            baseaddr = 0
        else:
            # This is a PIC binary
            # Relative symbols will be relative to the load address
            baseaddr = self.address

        dynsyms = set(elf.dynamic_symbols)
        idx = 0
        dynamic = True
        for s in elf.symbols:
            # Lief lets you access dynamic symbols separately,
            # but you need to access static symbols through
            # the list of all symbols
            if dynamic and s not in dynsyms:
                idx = 0
                dynamic = False

            # Build a symbol
            sym = ElfSymbol(
                idx=idx,
                dynamic=dynamic,
                name=s.name,
                type=s.type.value,
                bind=s.binding.value,
                visibility=s.visibility.value,
                shndx=s.shndx,
                value=s.value,
                size=s.size,
                baseaddr=baseaddr,
                defined=(s.shndx != 0),
            )
            # Save the sym, and temporarily tie it to its lief partner
            if dynamic:
                self._dynamic_symbols.append(sym)
            else:
                self._static_symbols.append(sym)

            self._syms_by_name.setdefault(sym.name, list()).append(sym)
            lief_to_elf[s] = sym

        if (
            self.platform.architecture == Architecture.MIPS32
            or self.platform.architecture == Architecture.MIPS64
        ):
            # All MIPS dynamic symbols have an implicit rela.
            # MIPS dynamic symbols always have a GOT entry;
            # to save space, the ABI just assumes that the rela exists
            if (
                DT_MIPS_GOTSYM not in self._dtags
                or DT_MIPS_LOCAL_GOTNO not in self._dtags
                or DT_PLTGOT not in self._dtags
            ):
                # GOT is missing... not sure what's up.
                log.error("MIPS binary missing GOT information")
            else:
                # We found the GOT info; we're actually a dynamic binary
                # Figure out the GOT entry size based on arch
                gotsym = self._dtags[DT_MIPS_GOTSYM]
                local_gotno = self._dtags[DT_MIPS_LOCAL_GOTNO]
                gotoff = self._dtags[DT_PLTGOT]

                if self.platform.architecture == Architecture.MIPS32:
                    gotent = 4
                    rela_type = R_MIPS_32
                else:
                    gotent = 8
                    rela_type = R_MIPS_64

                # Rebase the GOT offset relative to the image
                gotoff = self._rebase_file(gotoff)

                # Skip the first local_gotno entries
                gotoff += gotent * local_gotno

                for s in list(elf.dynamic_symbols)[gotsym:]:
                    sym = lief_to_elf[s]
                    rela = ElfRela(offset=gotoff, type=rela_type, symbol=sym, addend=0)
                    sym.relas.append(rela)
                    self._dynamic_relas.append(rela)

                    gotoff += gotent

        for r in list(elf.dynamic_relocations) + list(elf.pltgot_relocations):
            # Build a rela, and tie it to its symbol
            # TODO: some relas have no symbols.  Live with it.
            if r.symbol is None:
                continue
            sym = lief_to_elf[r.symbol]
            # Lief royally screws up relocation entries
            r_type = r.r_info(elf.header.identity_class)
            if elf.header.identity_class.value == 1:
                r_type &= 0xFF
            else:
                r_type &= 0xFFFFFFFF
            rela = ElfRela(
                offset=r.address + baseaddr, type=r_type, symbol=sym, addend=r.addend
            )
            sym.relas.append(rela)
            self._dynamic_relas.append(rela)

        for r in elf.object_relocations:
            if r.symbol is None:
                continue
            sym = lief_to_elf[r.symbol]
            # Lief royally screws up relocation entries
            r_type = r.r_info(elf.header.identity_class)
            if elf.header.identity_class.value == 1:
                r_type &= 0xFF
            else:
                r_type &= 0xFFFFFFFF
            rela = ElfRela(
                offset=r.address + baseaddr,
                type=r.type & 0xFFFF,
                symbol=sym,
                addend=r.addend,
            )
            sym.relas.append(rela)
            self._static_relas.append(rela)

    def _get_symbols(
        self, name: typing.Union[str, int], dynamic: bool
    ) -> typing.List[ElfSymbol]:
        if isinstance(name, str):
            # Caller wants to look up a symbol by name
            if name not in self._syms_by_name:
                raise ConfigurationError(f"No symbol named {name}")

            syms = self._syms_by_name[name]
            return list(syms)
        elif isinstance(name, int):
            if dynamic:
                return [self._dynamic_symbols[name]]
            else:
                return [self._static_symbols[name]]
        else:
            raise TypeError("Symbols must be specified by str names or int indexes")

    def get_symbol_value(
        self, name: typing.Union[str, int], dynamic: bool = False, rebase: bool = True
    ) -> int:
        """Get the value for a symbol

        The value of a symbol is usually an address or offset,
        although the precise meaning can vary a little.

        Arguments:
            name: The name of the symbol, or its index into the symbol table
            dynamic: If specified by index, whether to look in the static or dynamic symbol table
            rebase: Whether the recorded value is relative to the base address of this ELF.

        Returns:
            The integer value of this symbol
        """
        syms = self._get_symbols(name, dynamic)
        if len(syms) > 1:
            for sym in syms:
                if sym.value != syms[0].value and sym.baseaddr != syms[0].baseaddr:
                    raise ConfigurationError(f"Conflicting syms named {name}")

        val = syms[0].value
        if rebase:
            val += syms[0].baseaddr
        return val

    def get_symbol_size(self, name: typing.Union[str, int], dynamic: bool = False):
        """Get the size for a symbol

        If a symbol references a function or a data structure,
        this will hold its size in bytes

        Arguments:
            name: The name of the symbol, or its index into the symbol table
            dynamic: If specified by index, whether to look in the static or dynamic symbol table
            rebase: Whether the recorded value is relative to the base address of this ELF.

        Returns:
            The size of this symbol
        """
        syms = self._get_symbols(name, dynamic)
        if len(syms) > 1:
            for sym in syms:
                if sym.size != syms[0].size:
                    raise ConfigurationError(f"Conflicting syms named {name}")
        return syms[0].size

    def update_symbol_value(
        self,
        name: typing.Union[str, int, ElfSymbol],
        value: int,
        dynamic: bool = False,
        rebase: bool = True,
    ) -> None:
        """Update the value of a symbol

        This will alter the value of the symbol,
        and also propagate that updated value according to any associated relocations.

        Arguments:
            name: The name of the symbol, its index into the symbol table, or a specific symbol object
            value: The new value for the symbol
            dynamic: If specified by index, whether to look in the static or dynamic symbol table
            rebase: Whether the recorded value is relative to the base address of this ELF.
        """

        if isinstance(name, ElfSymbol):
            sym = name
        else:
            syms = self._get_symbols(name, dynamic)
            if len(syms) > 1:
                raise ConfigurationError(f"Multiple syms named {name}")
            sym = syms[0]

        if rebase:
            # Value provided is absolute; rebase it to the symbol's base address
            value -= sym.baseaddr

        # Update the value
        sym.value = value

        # Mark this symbol as defined
        sym.defined = True

        if self._relocator is not None:
            for rela in sym.relas:
                # Relocate!
                log.info(f"Relocating {rela}")
                self._relocator.relocate(self, rela)
        else:
            log.error(f"No platform defined; cannot relocate {name}!")

    def link_elf(self, elf: "ElfExecutable", dynamic: bool = True) -> None:
        """Link one ELF against another

        This roughly mimics the ELF linker;
        it looks for undefined symbols in the current file,
        and tries to populate their values from matching defined symbols
        from another file.

        Arguments:
            elf: The ELF from which to draw symbol values
            dynamic: Whether to link static or dynamic symbols
        """
        if dynamic:
            # Relocate rela.dyn and rela.plt
            relas = self._dynamic_relas
        else:
            # relocate static relocations
            relas = self._static_relas

        for rela in relas:
            my_sym = rela.symbol
            if my_sym.name == "":
                # This isn't a real symbol
                continue
            if my_sym.defined:
                # This is a defined symbol
                continue

            try:
                o_syms = elf._get_symbols(my_sym.name, dynamic)
            except ConfigurationError:
                continue

            o_syms = list(filter(lambda x: x.defined, o_syms))
            if len(o_syms) == 0:
                continue

            if len(set(map(lambda x: x.value, o_syms))) > 1:
                # Catch multiple symbols
                # If they all have the same value, we don't care.
                log.warning(f"Multiple symbols for {my_sym.name}")
                continue

            o_sym = o_syms[0]
            self.update_symbol_value(my_sym, o_sym.value + o_sym.baseaddr, rebase=True)

    def __getstate__(self):
        # Override the default pickling mechanism.
        # We can't save any lief data through a copy.
        state = self.__dict__.copy()
        state["_elf"] = None
        return state


__all__ = ["ElfExecutable"]



from .aarch64 import AArch64
from .amd64 import AMD64, AMD64AVX512
from .arm import ARMv5T, ARMv6M, ARMv6MThumb, ARMv7A, ARMv7M, ARMv7R
from .i386 import I386
from .loongarch import LoongArch64
from .mips import MIPS32BE, MIPS32EL
from .mips64 import MIPS64BE, MIPS64EL
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef
from .powerpc import PowerPC32, PowerPC64
from .riscv import RiscV64
from .xtensa import Xtensa

__all__ = [
    "AArch64",
    "AMD64",
    "AMD64AVX512",
    "ARMv5T",
    "ARMv6M",
    "ARMv6MThumb",
    "ARMv7M",
    "ARMv7R",
    "ARMv7A",
    "I386",
    "LoongArch64",
    "MIPS32EL",
    "MIPS32BE",
    "MIPS64EL",
    "MIPS64BE",
    "PlatformDef",
    "PowerPC32",
    "PowerPC64",
    "RegisterDef",
    "RegisterAliasDef",
    "RiscV64",
    "Xtensa",
]



import copy
import logging

from unicorn import unicorn_const

from ... import emulators, exceptions, hinting, instructions, state
from .. import analysis

logger = logging.getLogger(__name__)


class PointerFinder(analysis.Analysis):
    """A simple analysis that logs when a register is used as a pointer.

    Arguments:
        num_instructions: The number of instructions to execute.
    """

    def __init__(self, *args, num_instructions: int = 10, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_instructions = num_instructions

    name = "pointer-finder"
    description = ""
    version = "0.0.1"

    def find_the_pointer(self, cs_instruction, write):
        i = instructions.Instruction.from_capstone(cs_instruction)
        p = None
        if write:
            for w in i.writes:
                if type(w) is instructions.BSIDMemoryReferenceOperand:
                    p = w
                    break
        else:
            for r in i.reads:
                if type(r) is instructions.BSIDMemoryReferenceOperand:
                    p = r
                    break

        assert p, "we can't find the pointer"
        hint = hinting.PointerHint(message="Pointer Found", instruction=i, pointer=r)
        self.hinter.send(hint)

    def run(self, state: state.Machine) -> None:
        machine = copy.deepcopy(state)
        cpu = machine.get_cpu()
        emulator = emulators.UnicornEmulator(cpu.platform)
        machine.apply(emulator)

        def hook_valid_access(uc, access, address, size, value, user_data):
            instruction = emulator.current_instruction()
            if access == unicorn_const.UC_MEM_WRITE:
                self.find_the_pointer(instruction, True)
            else:
                self.find_the_pointer(instruction, False)

        def hook_invalid_access(uc, access, address, size, value, user_data):
            instruction = emulator.current_instruction()
            if access == unicorn_const.UC_MEM_WRITE_UNMAPPED:
                self.find_the_pointer(instruction, True)
            else:
                self.find_the_pointer(instruction, False)
            return False

        emulator.engine.hook_add(
            unicorn_const.UC_HOOK_MEM_WRITE | unicorn_const.UC_HOOK_MEM_READ,
            hook_valid_access,
        )
        emulator.engine.hook_add(
            unicorn_const.UC_HOOK_MEM_READ_UNMAPPED
            | unicorn_const.UC_HOOK_MEM_WRITE_UNMAPPED,
            hook_invalid_access,
        )

        for i in range(self.num_instructions):
            try:
                emulator.step()
            except exceptions.EmulationStop:
                break
            except exceptions.EmulationError as e:
                exhint = hinting.EmulationException(
                    message="Emulation single step raised an exception",
                    pc=emulator.current_instruction().address,
                    instruction_num=i,
                    exception=str(e),
                )
                self.hinter.send(exhint)
                break



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import AArch64SysVModel


class AArch64SysVFclose(Fclose, AArch64SysVModel):
    pass


class AArch64SysVFeof(Feof, AArch64SysVModel):
    pass


class AArch64SysVFerror(Ferror, AArch64SysVModel):
    pass


class AArch64SysVFgetc(Fgetc, AArch64SysVModel):
    pass


class AArch64SysVFgets(Fgets, AArch64SysVModel):
    pass


class AArch64SysVFopen(Fopen, AArch64SysVModel):
    pass


class AArch64SysVFprintf(Fprintf, AArch64SysVModel):
    pass


class AArch64SysVFputc(Fputc, AArch64SysVModel):
    pass


class AArch64SysVFputs(Fputs, AArch64SysVModel):
    pass


class AArch64SysVFread(Fread, AArch64SysVModel):
    pass


class AArch64SysVFscanf(Fscanf, AArch64SysVModel):
    pass


class AArch64SysVFseek(Fseek, AArch64SysVModel):
    pass


class AArch64SysVFtell(Ftell, AArch64SysVModel):
    pass


class AArch64SysVFwrite(Fwrite, AArch64SysVModel):
    pass


class AArch64SysVGetc(Getc, AArch64SysVModel):
    pass


class AArch64SysVGetchar(Getchar, AArch64SysVModel):
    pass


class AArch64SysVGets(Gets, AArch64SysVModel):
    pass


class AArch64SysVPrintf(Printf, AArch64SysVModel):
    pass


class AArch64SysVPutc(Putc, AArch64SysVModel):
    pass


class AArch64SysVPutchar(Putchar, AArch64SysVModel):
    pass


class AArch64SysVPuts(Puts, AArch64SysVModel):
    pass


class AArch64SysVRemove(Remove, AArch64SysVModel):
    pass


class AArch64SysVRename(Rename, AArch64SysVModel):
    pass


class AArch64SysVRewind(Rewind, AArch64SysVModel):
    pass


class AArch64SysVScanf(Scanf, AArch64SysVModel):
    pass


class AArch64SysVSnprintf(Snprintf, AArch64SysVModel):
    pass


class AArch64SysVSprintf(Sprintf, AArch64SysVModel):
    pass


class AArch64SysVSscanf(Sscanf, AArch64SysVModel):
    pass


class AArch64SysVClearerr(Clearerr, AArch64SysVModel):
    pass


class AArch64SysVFflush(Fflush, AArch64SysVModel):
    pass


class AArch64SysVFreopen(Freopen, AArch64SysVModel):
    pass


class AArch64SysVFgetpos(Fgetpos, AArch64SysVModel):
    pass


class AArch64SysVFsetpos(Fsetpos, AArch64SysVModel):
    pass


class AArch64SysVTmpfile(Tmpfile, AArch64SysVModel):
    pass


class AArch64SysVTmpnam(Tmpnam, AArch64SysVModel):
    pass


class AArch64SysVUngetc(Ungetc, AArch64SysVModel):
    pass


class AArch64SysVVfprintf(Vfprintf, AArch64SysVModel):
    pass


class AArch64SysVVfscanf(Vfscanf, AArch64SysVModel):
    pass


class AArch64SysVVprintf(Vprintf, AArch64SysVModel):
    pass


class AArch64SysVVscanf(Vscanf, AArch64SysVModel):
    pass


class AArch64SysVVsnprintf(Vsnprintf, AArch64SysVModel):
    pass


class AArch64SysVVsprintf(Vsprintf, AArch64SysVModel):
    pass


class AArch64SysVVsscanf(Vsscanf, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVFclose",
    "AArch64SysVFeof",
    "AArch64SysVFerror",
    "AArch64SysVFgetc",
    "AArch64SysVFgets",
    "AArch64SysVFopen",
    "AArch64SysVFprintf",
    "AArch64SysVFputc",
    "AArch64SysVFputs",
    "AArch64SysVFread",
    "AArch64SysVFscanf",
    "AArch64SysVFseek",
    "AArch64SysVFtell",
    "AArch64SysVFwrite",
    "AArch64SysVGetc",
    "AArch64SysVGetchar",
    "AArch64SysVGets",
    "AArch64SysVPrintf",
    "AArch64SysVPutc",
    "AArch64SysVPutchar",
    "AArch64SysVPuts",
    "AArch64SysVRemove",
    "AArch64SysVRename",
    "AArch64SysVRewind",
    "AArch64SysVScanf",
    "AArch64SysVSnprintf",
    "AArch64SysVSprintf",
    "AArch64SysVSscanf",
    "AArch64SysVClearerr",
    "AArch64SysVFflush",
    "AArch64SysVFreopen",
    "AArch64SysVFgetpos",
    "AArch64SysVFsetpos",
    "AArch64SysVTmpfile",
    "AArch64SysVTmpnam",
    "AArch64SysVUngetc",
    "AArch64SysVVfprintf",
    "AArch64SysVVfscanf",
    "AArch64SysVVprintf",
    "AArch64SysVVscanf",
    "AArch64SysVVsnprintf",
    "AArch64SysVVsprintf",
    "AArch64SysVVsscanf",
]



import logging

log = logging.getLogger(__name__)


class TUIContinueException(Exception):
    """Exception for signaling that a TUI handler shouldn't exit the loop."""

    pass


class SimpleTUI:
    def __init__(self, help_banner="Available commands:"):
        self._cases = dict()
        self._shorts = dict()
        self._hints = dict()
        self._help_banner = help_banner
        self.add_case("help", self.print_help, hint="Print this message")

    def add_case(self, name, handler, hint=None):
        name = name.lower()
        short = name[0]
        if short in self._shorts:
            raise ValueError(
                "Colliding short name {short}: already bound for {self._shorts[short]}"
            )
        if name in self._cases:
            raise ValueError("Case {name} already registered")
        self._cases[name] = handler
        self._shorts[short] = name
        self._hints[name] = hint

    def print_help(self, **kwargs):
        log.warn(self._help_banner)
        for name in self._cases:
            log.warn(
                f'- {name} | {name[0]}:\t\t\t{self._hints[name] if self._hints[name] is not None else ""}'
            )
        raise TUIContinueException()

    def handle(self, _default, _disabled, **kwargs):
        prompt = " | ".join(
            map(
                lambda x: x.upper() if x == _default else x,
                filter(lambda x: x not in _disabled, self._cases.keys()),
            )
        )
        prompt = f"( {prompt} ) > "
        while True:
            opt = input(prompt).lower()
            if opt == "":
                if _default is None:
                    log.error("No default case available")
                    continue
                opt = _default

            if opt in self._shorts:
                opt = self._shorts[opt]

            if opt in _disabled:
                log.error("Option {opt} not available")
                continue

            if opt not in self._cases:
                log.error("Unknown option {opt}")
                continue

            try:
                return self._cases[opt](**kwargs)
            except TUIContinueException:
                continue



import logging
import sys


class CharacterLevelFilter(logging.Filter):
    """Adds logging level as a single character for formatting."""

    CHARACTERS = {
        logging.DEBUG: "-",
        logging.INFO: "+",
        logging.WARNING: "!",
        logging.ERROR: "*",
        logging.CRITICAL: "#",
    }

    def filter(self, record):
        record.levelchar = self.CHARACTERS.get(record.levelno, " ")
        return True


class ColorLevelFilter(logging.Filter):
    """Adds logging level as a color for formatting."""

    WHITE_DIM = "\x1b[37;2m"
    WHITE = "\x1b[37m"
    NORMAL_DIM = "\x1b[0;2m"
    NORMAL = "\x1b[0m"
    YELLOW = "\x1b[33m"
    RED = "\x1b[31m"
    RED_BOLD = "\x1b[31;1m"
    END = "\x1b[0m"
    NULL = END

    COLORS = {
        logging.DEBUG: NORMAL_DIM,
        logging.INFO: NORMAL,
        logging.WARNING: YELLOW,
        logging.ERROR: RED,
        logging.CRITICAL: RED_BOLD,
    }

    def filter(self, record):
        record.levelcolor = self.COLORS.get(record.levelno, self.NULL)
        return True


def setup_logging(
    level: int = logging.INFO,
    verbose: bool = False,
    colors: bool = True,
    clear: bool = True,
) -> None:
    """Setup log handling.

    Note: this should only be called once.

    Arguments:
        level: Logging level (from `logging` module).
        verbose: Enable verbose logging mode.
        colors: Enable logging colors (if supported).
        clear: Remove all other handlers from the root logger.
    """

    if verbose:
        format = "[%(asctime)s %(levelname)-7s|%(name)s]: %(message)s"
    else:
        format = "[%(levelchar)s] %(message)s"

    if colors and sys.stderr.isatty():
        format = f"%(levelcolor)s{format}{ColorLevelFilter.END}"

    root = logging.getLogger()
    root.setLevel(level)

    if clear:
        handlers = list(root.handlers)
        for handler in handlers:
            root.removeHandler(handler)

    formatter = logging.Formatter(format)

    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.addFilter(CharacterLevelFilter())
    handler.addFilter(ColorLevelFilter())

    handler.setFormatter(formatter)

    root.addHandler(handler)


__all__ = ["setup_logging"]
